from profit.util.halton import halton as _halton_base
from profit.util import check_ndim
import numpy as np
from profit.util.base_class import CustomABC


# TODO: Sample all variables from halton.
EXCLUDE_FROM_HALTON = ('output', 'constant', 'uniform', 'loguniform', 'normal', 'linear', 'independent')


def halton(size=(1, 1)):
    if isinstance(size, (tuple, list, np.ndarray)):
        return _halton_base(*size)
    else:
        return check_ndim(_halton_base(size, 1))


def uniform(start=0, end=1, size=None):
    return check_ndim(start + np.random.random(size) * (end - start))


def loguniform(start=1e-6, end=1, size=None):
    return check_ndim(start * np.exp((np.log(end) - np.log(start)) * np.random.random(size)))


def normal(mu=0, std=1, size=None):
    return check_ndim(np.random.normal(mu, std, size))


def linear(start=0, end=1, size=1):
    return check_ndim(np.linspace(start, end, size))


def independent(start=0, end=1, size=1):
    return linear(start, end, size)

def constant(value=0, size=None):
    return check_ndim(np.full(size, value))


class VariableGroup:
    """Table of input, output and independent variables.

    Parameters:
        samples (int): Samples of the training data.

    Attributes:
        samples (int): Samples of the training data.
        list (list): List of all variables in the order of the user entry.
    """

    def __init__(self, samples):
        self.samples = samples
        self.list = []

    @property
    def all(self):
        """
        Returns:
            View on all variables.

        Not working yet for vector output.
        """
        values = np.hstack([v.value for v in self.list])
        dtypes = [(v.name, v.dtype) for v in self.list]
        return values.view(dtype=dtypes)

    @property
    def as_dict(self):
        """
        Returns:
            View of all variables as a dictionary.
        """
        input_dict = {k: v.as_dict() for k, v in self.input_dict.items()}
        independent_dict = {v.name: v.as_dict() for v in self.list if v.__class__ == IndependentVariable}
        output_dict = {k: v.as_dict() for k, v in self.output_dict.items()}
        return {**input_dict, **independent_dict, **output_dict}

    @property
    def input(self):
        """
        Returns:
            View of the input variables only. Also excluded are independent variables.
        """
        return np.hstack([v.value for v in self.list if v.__class__ in (InputVariable, ActiveLearningVariable)])

    @property
    def named_input(self):
        """
        Returns:
            Ndarray with dtype of the input variables.
        """
        dtypes = [(v.name, v.dtype) for v in self.list if v.__class__ in (InputVariable, ActiveLearningVariable)]
        return np.rec.fromarrays([v.value for v in self.list
                                  if v.__class__ in (InputVariable, ActiveLearningVariable)], dtype=dtypes)

    @property
    def input_dict(self):
        """
        Returns:
            Dictionary of the input variables.
        """
        return {v.name: v for v in self.list if v.__class__ in (InputVariable, ActiveLearningVariable)}

    @property
    def input_list(self):
        """
        Returns:
            List of input variables without independent variables.
        """
        return [v for v in self.list if v.__class__ in (InputVariable, ActiveLearningVariable)]

    @property
    def kind_dict(self):
        kinds = {}
        for v in self.input_list:
            if v["kind"] in kinds:
                kinds[v["kind"]].append(v)
            else:
                kinds[v["kind"]] = [v]
        return kinds

    @property
    def output(self):
        """
        Returns:
            View on the output variables only.
        """
        return np.hstack([v.value for v in self.list if v.__class__ == OutputVariable])

    @property
    def named_output(self):
        """
        Returns:
            Ndarray with dtype of the output variables.
        """
        dtypes = [(v.name, v.dtype) for v in self.list if v.__class__ == OutputVariable]
        return self.output.view(dtype=dtypes)

    @property
    def formatted_output(self):
        dtype = []
        columns = {}
        # prepare dtype
        for key in self.output_dict:
            spec = self[key]
            if spec.size[-1] == 1:
                dtype.append((key, spec.dtype))
                columns[key] = [key]
            else:
                ranges = []
                columns[key] = []
                for dep in spec.dependent:
                    ranges.append(dep.value)
                meshes = [m.flatten() for m in np.meshgrid(*ranges)]
                for i in range(meshes[0].size):
                    name = key + '(' + ', '.join([f'{m[i]}' for m in meshes]) + ')'
                    dtype.append((name, spec.dtype))
                    columns[key].append(name)
        # fill data
        output = np.zeros(len(self.named_output), dtype=dtype)
        for key, spec in self.output_dict.items():
            if spec.size[-1] == 1:
                output[key] = self.named_output[key].flatten()
            else:
                for i, values in enumerate(self.named_output):
                    output[columns[key]][i] = tuple(values[key])
        output = output.reshape(-1, 1)
        return output

    @property
    def output_dict(self):
        """
        Returns:
            Dictionary of the output variables.
        """
        return {v.name: v for v in self.list if v.__class__ == OutputVariable}

    @property
    def output_list(self):
        """
        Returns:
            List of output variables.
        """
        return [v for v in self.list if v.__class__ == OutputVariable]

    def __getitem__(self, item):
        """Implements dict like behavior to get a variable by its identifier or index.

        Parameters:
            item (int/str): Index or label of variable.
        Returns:
            Variable.
        """
        if isinstance(item, str):
            item = [i for i, v in enumerate(self.list) if item == v.name]
            if len(item) > 0:
                item = item[0]
        return self.list[item]

    def add(self, variables):
        """Adds a single or a list of variables to the table.
        If a list is added, a common n-D halton sequence is generated and the variables are transformed
        according to their distribution.

        Parameters:
            variables (Variable/list): Variable(s) to add.
        """
        if not isinstance(variables, (list, tuple)):
            variables = [variables]

        for v in variables:
            if isinstance(v, Variable):
                self.list.append(v)
            elif isinstance(v, dict):
                self.list.append(Variable.create(**v))
            else:
                raise NotImplementedError

        self.generate_from_halton()
        for v in self.list:
            if any(e in v.kind.lower() for e in EXCLUDE_FROM_HALTON) and v.__class__ != OutputVariable:
                v.generate_values()

        for v in self.list:
            if v.__class__ == OutputVariable:
                dep = [v if isinstance(v, str) else v['name'] for v in v.dependent]
                ind = [i for i in self.list if i.__class__ == IndependentVariable and i.name in dep]
                if ind:
                    v.resolve_dependent(ind)

    def delete_variable(self, columns):
        """Deletes one or more variables from the table.

        Parameters:
            columns (int/list): Columns of the table to remove.
        """
        if not isinstance(columns, (list, tuple)):
            columns = [columns]
        for col in columns:
            if isinstance(col, str):
                col = [i for i, v in enumerate(self.list) if v.name == col][0]
            self.list.pop(col)

    def delete_sample(self, rows):
        """Deletes one or more rows of the table.

        Parameters:
            rows (int/list): Rows to delete.
        """
        if not isinstance(rows, (list, tuple)):
            rows = [rows]
        for v in self.list:
            v.value = np.delete(v.value, rows, axis=0)

    def generate_from_halton(self):
        """Generates a common halton sequence for all variables where this is possible and transforms them
        according to their distribution."""
        halton_variables = [v for v in self.list if v.kind.lower() not in EXCLUDE_FROM_HALTON]
        if halton_variables:
            nd_halton_seq = halton((self.samples, len(halton_variables)))
            for idx, v in enumerate(halton_variables):
                v.generate_values(nd_halton_seq[:, idx])


class Variable(CustomABC):
    """Base class for a single variable.
    To create input, independent and output variables, use the cls.create() or cls.create_from_str() methods.

    Attributes:
        name (str): Name of the variable.
        kind (str): Distribution for input variables, 'Output' or 'Independent'.
        size (tuple): Size as (nsamples, ndim).
        value (ndarray): Data.
        dtype (dtype): Datatype.
    """

    labels = {}

    def __init__(self, name, kind, size, value=None, dtype=np.float64):
        self.name = name
        self.kind = kind
        self.size = size
        self.value = value if value is not None else np.empty(size)
        assert self.value.shape == size
        self.dtype = dtype

    @property
    def named_value(self):
        """
        Returns:
            Ndarray with dtype.
        """
        return np.array(self.value, dtype=[(self.name, self.dtype)])

    @classmethod
    def create_from_str(cls, name, size, v_str):
        """Creates a Variable instance from a string. E.g. 'Uniform(3.4, 7.8)'

        Parameters:
            name (str): Name of the variable.
            size (tuple): Size as (nsamples, ndim).
            v_str (str): String from which the variable is constructed.

        Returns:
            Variable.
        """
        from re import split

        def try_parse(s):
            funcs = [int, float]
            for f in funcs:
                try:
                    return f(s)
                except ValueError:
                    pass
            return s

        if isinstance(try_parse(v_str), (int, float)):
            v_str = 'Constant({})'.format(try_parse(v_str))

        parsed = split('[()]', v_str)
        kind = parsed[0]
        args = parsed[1] if len(parsed) >= 2 else ''
        entries = tuple(try_parse(a) for a in args.split(',')) if args != '' else tuple()
        entries_dict = cls.labels.get(kind.lower(), cls.labels['input']).parse_entries(entries) \
            if len(entries) > 0 else {}

        dtype = type(entries[0]) if kind.lower() == 'constant' else np.float64

        v_dict = {'name': name, 'kind': kind, 'size': size, 'dtype': dtype}
        v_dict = {**v_dict, **entries_dict}
        return cls.create(**v_dict)

    @classmethod
    def create(cls, name, kind, size, value=None, dtype=np.float64, **kwargs):
        """Directly creates a variable from keyword entries.

        Parameters:
            name (str): Name of the variable.
            kind (str): Distribution of input variables, 'Output' or 'Independent'.
            size (tuple): Size as (nsamples, ndim).
            kwargs (tuple/str): Keyword arguments depending on the sub variables.
                E.g. constraints for input variables, a search distribution for active learning variables or
                dependent variables of outputs.
            value (ndarray): Data.
            dtype (dtype): Datatype.

        Returns:
            Variable.
        """
        if isinstance(dtype, str) and kind.lower() != 'constant':
            dtype = np.dtype(dtype).type
        return cls.labels.get(kind.lower(), cls.labels['input'])(name=name, kind=kind, size=size,
                                                                 value=value, dtype=dtype, **kwargs)

    def as_dict(self):
        """Dictionary of the variable attributes."""
        return {k: v for k, v in vars(self).items()}

    def __getitem__(self, item):
        """Implement dict like behavior to get an attribute by using square brackets.

        Paramters:
            item (str)

        Returns:
            Attribute
        """
        return getattr(self, item)


@Variable.register("input")
class InputVariable(Variable):
    """Sub class for input variables."""

    def __init__(self, name, kind, size, constraints=(0, 1), value=None, dtype=np.float64):
        super().__init__(name, kind, size, value, dtype)
        self.constraints = constraints

    def generate_values(self, halton_seq=None):
        if halton_seq is None:
            if len(self.constraints) == 3:
                self.value = globals().get(self.kind.lower())(*self.constraints).astype(self.dtype)
            else:
                self.value = globals().get(self.kind.lower())(*self.constraints, size=self.size).astype(self.dtype)
        else:
            self.value = check_ndim((self.constraints[1] - self.constraints[0]) * halton_seq + self.constraints[0]).astype(self.dtype)

    @classmethod
    def parse_entries(cls, entries):
        return {'constraints': entries}

    def create_Xpred(self, size):
        if not isinstance(size, tuple):
            size = (size, 1)
        if 'log' in self.kind.lower():
            return self.constraints[0] * np.exp((np.log(self.constraints[1]) - np.log(self.constraints[0]))
                                                           * np.linspace(0, 1, size[0])).reshape(size)
        elif 'constant' in self.kind.lower():
            return self.value[0]
        else:
            return np.linspace(*self.constraints, size[0]).reshape(size)


@Variable.register("independent")
class IndependentVariable(InputVariable):
    """Sub class for independent variables."""

    def __init__(self, name, kind, size, constraints=(0, 1), value=None, dtype=np.float64):
        super().__init__(name, kind, size, constraints, value, dtype)
        if value is None:
            self.generate_values()
            self.size = self.value.shape


@Variable.register("activelearning")
class ActiveLearningVariable(InputVariable):
    """Sub class for active learning variables."""

    def __init__(self, name, kind, size, distr='uniform', constraints=(0, 1), value=None, dtype=np.float64):
        if value is None:
            value = np.full(size, np.nan)
        super().__init__(name, kind, size, constraints, value, dtype)
        self.distr = distr

    @classmethod
    def parse_entries(cls, entries):
        return {'constraints': entries[:2], 'distr': entries[2].strip() if len(entries) == 3 else 'uniform'}

    def generate_values(self, halton_seq=None):
        return check_ndim(np.full(self.size, np.nan))

    def create_Xpred(self, size):
        if not isinstance(size, tuple):
            size = (size, 1)
        if self.distr.lower() == 'log':
            return self.constraints[0] * np.exp((np.log(self.constraints[1]) - np.log(self.constraints[0]))
                                                           * np.linspace(0, 1, size[0])).reshape(size)
        else:
            return np.linspace(*self.constraints, size[0]).reshape(size)


@Variable.register("output")
class OutputVariable(Variable):
    """Sub class for output variables."""

    def __init__(self, name, kind, size, dependent=(), value=None, dtype=np.float64):
        super().__init__(name, kind, size, value, dtype)
        if not isinstance(dependent, (list, tuple)):
            dependent = [dependent]
        self.dependent = dependent
        self.value = value if value is not None else np.full(self.size, np.nan)

    @classmethod
    def parse_entries(cls, entries):
        return {'dependent': entries}

    def resolve_dependent(self, ind):
        """Create a :class:`.Variable` instance for the independent variables of vector outputs
        and set :attr:`dependent`.

        Parameters:
            ind (profit.util.variable.IndependentVariable or list[profit.util.variable.IndependentVariable]):
                Independent variables.
        """
        if not isinstance(ind, (list, tuple)):
            ind = [ind]

        dvars = []
        for d in self.dependent:
            if isinstance(d, str):
                d = {'name': d}
            dv = [v for v in ind if v.name == d['name']][0]
            self.size = (self.size[0], dv.value.shape[0])
            self.value = np.full(self.size, np.nan)
            dvars.append(dv)
        self.dependent = dvars

    def as_dict(self):
        return {k: v if k != 'dependent' else [vi.as_dict() for vi in v] for k, v in vars(self).items()}



"""
Abstract base class for all surrogate models.

Class structure:
- Surrogate
    - GaussianProcess
        - GPSurrogate (Custom)
        - GPySurrogate (GPy)
        - SklearnGPSurrogate (Sklearn)
    - ANN
        - ANNSurrogate (Pytorch)
        - Autoencoder (Pytorch)
    - LinearRegression
        - Linear regression surrogates (Work in progress)
"""

from abc import abstractmethod
import numpy as np
from profit.util.base_class import CustomABC
from profit.defaults import fit as defaults


class Surrogate(CustomABC):
    """Base class for all surrogate models.

    Attributes:
        trained (bool): Flag that indicates if the model is already trained and ready to make predictions.
        fixed_sigma_n (bool): Indicates if the data noise should be optimized or not.
        Xtrain (ndarray): Input training points.
        ytrain (ndarray): Observed output data.
            Vector output is supported for independent variables only.
        ndim (int): Dimension of input data.
        output_ndim (int): Dimension of output data.
        input_encoders (list of profit.sur.encoders.Encoder): Encoding used on input data.
        output_encoders (list of profit.sur.encoders.Encoder): Encoding used on output data.

    Default parameters:
        surrogate: GPy
        save: ./model_{surrogate label}.hdf5
        load: False
        fixed_sigma_n: False
        input_encoders: [{'class': 'exclude', 'columns': {constant columns}
                         {'class': 'log10', 'columns': {log input columns}, 'parameters': {}},
                         {'class': 'normalization', 'columns': {input columns}, 'parameters': {}}]
        output_encoders: [{'class': 'normalization', 'columns': {output columns}, 'parameters': {}}]
    """

    labels = {}  # All surrogates are registered here

    def __init__(self):
        self.trained = False
        self.fixed_sigma_n = False

        self.Xtrain = None
        self.ytrain = None
        self.ndim = None  # TODO: Consistency between len(base_config['input']) and self.Xtrain.shape[-1]
        self.output_ndim = 1

        self.input_encoders = []
        self.output_encoders = []

    def encode_training_data(self):
        """Encodes the input and output training data.
        """
        for enc in self.input_encoders:
            self.Xtrain = enc.encode(self.Xtrain)
        for enc in self.output_encoders:
            self.ytrain = enc.encode(self.ytrain)

    def decode_training_data(self):
        """Applies the decoding function of the encoder in reverse order on the input and output training data.
        """
        for enc in self.input_encoders[::-1]:
            self.Xtrain = enc.decode(self.Xtrain)
        for enc in self.output_encoders[::-1]:
            self.ytrain = enc.decode(self.ytrain)

    def encode_predict_data(self, x):
        """Transforms the input prediction points according to the encoder used for training.

        Parameters:
            x (ndarray): Prediction input points.

        Returns:
            ndarray: Encoded and normalized prediction points.
        """
        for enc in self.input_encoders:
            x = enc.encode(x)
        return x

    def decode_predict_data(self, ym, yv):
        """Rescales and then back-transforms the predicted output.

        Parameters:
            ym (ndarray): Predictive output.
            yv (ndarray): Variance of predicted output.

        Returns:
            tuple: a tuple containing:
                - ym (ndarray) Rescaled and decoded output values at the test input points.
                - yv (ndarray): Rescaled predictive variance.
        """

        for enc in self.output_encoders[::-1]:
            ym = enc.decode(ym)
            yv = enc.decode_variance(yv)
        return ym, yv

    def add_input_encoder(self, encoder):
        """Add encoder on input data.

        Parameters:
            encoder (profit.sur.encoder.Encoder)
        """
        self.input_encoders.append(encoder)

    def add_output_encoder(self, encoder):
        """Add encoder on output data.

        Parameters:
            encoder (profit.sur.encoder.Encoder)
        """
        self.output_encoders.append(encoder)

    @abstractmethod
    def train(self, X, y, fixed_sigma_n=defaults['fixed_sigma_n']):
        r"""Trains the surrogate on input points X and model outputs y.

        Depending on the surrogate, the signature can vary.

        Parameters:
            X (ndarray): Input training points.
            y (ndarray): Observed output data.
            fixed_sigma_n (bool): Whether the noise $\sigma_n$ is fixed during optimization.
        """
        pass

    @abstractmethod
    def predict(self, Xpred, add_data_variance=True):
        r"""Predicts model output y for input Xpred based on surrogate.

        Parameters:
            Xpred (ndarray/list): Input points for prediction.
            add_data_variance (bool): Adds the data noise $\sigma_n^2$ to the prediction variance.
                This is especially useful for plotting.

        Returns:
            tuple: a tuple containing:
                - ymean (ndarray) Predicted output values at the test input points.
                - yvar (ndarray): Generally the uncertainty of the fit. For Gaussian Processes this is
                the diagonal of the posterior covariance matrix.
        """
        pass

    @abstractmethod
    def save_model(self, path):
        """Saves the surrogate to a file. The file format can vary between surrogates.
        As default, the surrogate is saved to 'base_dir/model_{surrogate_label}.hdf5'.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """
        pass

    @classmethod
    @abstractmethod
    def load_model(cls, path):
        """Loads a saved surrogate from a file. The file format can vary between surrogates.

        Identifies the surrogate by its class label in the file name.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            profit.sur.Surrogate: Instantiated surrogate model.
        """
        label = defaults['surrogate']
        for f in filter(lambda l: l in path, cls.labels):
            if len(f) > len(label):
                label = f
        return cls[label].load_model(path)

    @classmethod
    @abstractmethod
    def from_config(cls, config, base_config):
        """Instantiates a surrogate based on the parameters given in the configuration file and delegates to child.

        Parameters:
            config (dict): Only the 'fit' part of the base_config.
            base_config (dict): The whole configuration parameters.
        """
        from .encoders import Encoder

        child = cls[config['surrogate']]

        if config.get('load'):
            child_instance = child.load_model(config['load'])
        else:
            child_instance = child.from_config(config, base_config)
            # Set global attributes
            child_instance.ndim = len(base_config['input'])
            child_instance.output_ndim = len(base_config['output'])
            child_instance.fixed_sigma_n = config['fixed_sigma_n']
            for enc in config['_input_encoders']:
                child_instance.add_input_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))
            for enc in config['_output_encoders']:
                child_instance.add_output_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))
        return child_instance

    def plot(self, Xpred=None, independent=None, show=False, ref=None, add_data_variance=True, axes=None):
        r"""Simple plotting for dimensions <= 2.

        Fore more sophisticated plots use the command 'profit ui'.

        Parameters:
            Xpred (ndarray): Prediction points where the fit is plotted. If None, it is inferred from the
                training points.
            independent (dict): Dictionary of independent variables from config.
            show (bool): If the figure should be shown directly.
            ref (ndarray): Reference function which is fitted.
            add_data_variance (bool): Adds the data noise $\sigma_n^2$ to the prediction variance.
            axes (matplotlib.pyplot.axes): Axes object to insert the plot into. If None, a new figure is created.
        """

        import matplotlib.pyplot as plt
        if Xpred is None:
            Xpred = self.default_Xpred()
        ypred, yvarpred = self.predict(Xpred, add_data_variance=add_data_variance)
        ystd_pred = np.sqrt(yvarpred)
        if independent:
            # 2D with one input parameter and one independent variable.
            if self.ndim == 1 and ypred.ndim == 2:
                ax = axes or plt.axes(projection='3d')
                xind = np.hstack([v['value'] for v in independent.values()])
                xtgrid = np.meshgrid(*[xind, self.Xtrain])
                xgrid = np.meshgrid(*[xind, Xpred])
                for i in range(self.Xtrain.shape[0]):
                    ax.plot(xtgrid[0][i], xtgrid[1][i], self.ytrain[i], color='blue', linewidth=2)
                ax.plot_surface(xgrid[0], xgrid[1], ypred, color='red', alpha=0.8)
                ax.plot_surface(xgrid[0], xgrid[1], ypred + 2 * ystd_pred, color='grey', alpha=0.6)
                ax.plot_surface(xgrid[0], xgrid[1], ypred - 2 * ystd_pred, color='grey', alpha=0.6)
            else:
                raise NotImplementedError("Plotting is only implemented for dimensions <= 2. Use profit ui instead.")
        else:
            if self.ndim == 1 and ypred.shape[-1] == 1:
                # Only one input parameter to plot.
                ax = axes or plt.axes()
                if ref:
                    ax.plot(Xpred, ref(Xpred), color='red')
                ax.plot(Xpred, ypred)
                ax.scatter(self.Xtrain, self.ytrain, marker='x', s=50, c='k')
                ax.fill_between(Xpred.flatten(),
                                ypred.flatten() + 2 * ystd_pred.flatten(), ypred.flatten() - 2 * ystd_pred.flatten(),
                                color='grey', alpha=0.6)
            elif self.ndim == 2 and ypred.shape[-1] == 1:
                # Two fitted input variables.
                ax = axes or plt.axes(projection='3d')
                ypred = ypred.flatten()
                ystd_pred = ystd_pred.flatten()
                ax.scatter(self.Xtrain[:, 0], self.Xtrain[:, 1], self.ytrain, color='red', alpha=0.8)
                ax.plot_trisurf(Xpred[:, 0], Xpred[:, 1], ypred, color='red', alpha=0.8)
                ax.plot_trisurf(Xpred[:, 0], Xpred[:, 1], ypred + 2 * ystd_pred, color='grey', alpha=0.6)
                ax.plot_trisurf(Xpred[:, 0], Xpred[:, 1], ypred - 2 * ystd_pred, color='grey', alpha=0.6)
            elif self.ndim == 1 and self.output_ndim == 2:
                # One input variable and two outputs
                ax = axes or plt.axes()
                for d in range(self.output_ndim):
                    yp = ypred[:, d]
                    ystd_p = ystd_pred[:, d]
                    ax.scatter(self.Xtrain, self.ytrain[:, d], alpha=0.8)
                    ax.plot(Xpred, yp)
                    ax.fill_between(Xpred.flatten(), yp + 2 * ystd_p, yp - 2 * ystd_p, alpha=0.6)
            else:
                raise NotImplementedError("Plotting is only implemented for dimension <= 2. Use profit ui instead.")

        if show:
            plt.show()

    def default_Xpred(self):
        """Infer prediction values from training points in each dimension.

        Currently a dense grid is created. This becomes inefficient for > 3 dimensions.

        Returns:
            ndarray: Prediction points.
        """

        if self.ndim <= 3:
            minval = self.Xtrain.min(axis=0)
            maxval = self.Xtrain.max(axis=0)
            npoints = [50] * len(minval)
            xpred = [np.linspace(minv, maxv, n) for minv, maxv, n in zip(minval, maxval, npoints)]
            return np.hstack([xi.flatten().reshape(-1, 1) for xi in np.meshgrid(*xpred)])
        else:
            raise RuntimeError("Require x for prediction in > 3 dimensions!")



import numpy as np
from profit.run.worker import Worker
nt = 250


@Worker.wrap("mockup_mcmc", ['u', 'v'], 'f')
def f(u, v):
    t = np.linspace(0, 2 * (1 - 1 / nt), nt)
    y_model = u * np.sin((t - v)**3)
    return y_model



import numpy as np
from profit.sur import Surrogate
from profit.sur.gp import GaussianProcess
from profit.defaults import fit_gaussian_process as defaults, fit as base_defaults


@Surrogate.register('GPy')
class GPySurrogate(GaussianProcess):
    """Surrogate for https://github.com/SheffieldML/GPy.

    Attributes:
        model (GPy.models): Model object of GPy.
    """

    def __init__(self):
        import GPy
        self.GPy = GPy
        super().__init__()
        self.model = None

    def train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
              fixed_sigma_n=base_defaults['fixed_sigma_n']):
        self.pre_train(X, y, kernel, hyperparameters, fixed_sigma_n)
        self.model = self.GPy.models.GPRegression(self.Xtrain, self.ytrain, self.kernel,
                                                  noise_var=self.hyperparameters['sigma_n'] ** 2)
        self.optimize()
        self.post_train()

    def add_training_data(self, X, y):
        """Adds training points to the existing dataset.

        This is important for Active Learning. The data is added but the hyperparameters are not optimized yet.

        Parameters:
            X (ndarray): Input points to add.
            y (ndarray): Observed output to add.
        """
        self.Xtrain, self.ytrain = np.concatenate([self.Xtrain, X], axis=0), np.concatenate([self.ytrain, y], axis=0)
        self.encode_training_data()
        self.model.set_XY(self.Xtrain, self.ytrain)
        self.decode_training_data()

    def set_ytrain(self, y):
        """Set the observed training outputs. This is important for active learning.

        Parameters:
        y (np.array): Full training output data.
        """
        self.ytrain = np.atleast_2d(y.copy())
        self.encode_training_data()
        self.model.set_XY(self.Xtrain, self.ytrain)
        self.decode_training_data()

    def predict(self, Xpred, add_data_variance=True):
        Xpred = self.pre_predict(Xpred)
        ymean, yvar = self.model.predict(Xpred, include_likelihood=add_data_variance)
        ymean, yvar = self.decode_predict_data(ymean, yvar)
        return ymean, yvar

    def save_model(self, path):
        """Save the model as dict to a .hdf5 file.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """

        from profit.util.file_handler import FileHandler

        sur_dict = {attr: getattr(self, attr) for attr in ('Xtrain', 'ytrain')}
        sur_dict['model'] = self.model.to_dict()
        sur_dict['input_encoders'] = str([enc.repr for enc in self.input_encoders])
        sur_dict['output_encoders'] = str([enc.repr for enc in self.output_encoders])
        FileHandler.save(path, sur_dict)

    @classmethod
    def load_model(cls, path):
        """Loads a saved model from a .hdf5 file and updates its attributes. In case of a multi-output model, the .pkl
        file is loaded, since .hdf5 is not supported yet.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            GPy.models: Instantiated surrogate model.
        """

        from profit.util.file_handler import FileHandler
        from profit.sur.encoders import Encoder
        from GPy import models
        from numpy import array  # needed for eval of arrays

        self = cls()
        sur_dict = FileHandler.load(path, as_type='dict')
        self.model = models.GPRegression.from_dict(sur_dict['model'])
        self.Xtrain, self.ytrain = sur_dict['Xtrain'], sur_dict['ytrain']

        for enc in eval(sur_dict['input_encoders']):
            self.add_input_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))
        for enc in eval(sur_dict['output_encoders']):
            self.add_output_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))

        # Initialize the encoder by encoding and decoding the training data once.
        self.encode_training_data()
        self.decode_training_data()

        self.kernel = self.model.kern
        self.ndim = self.Xtrain.shape[-1]
        self._set_hyperparameters_from_model()
        self.trained = True
        self.print_hyperparameters("Loaded")
        return self

    def select_kernel(self, kernel):
        """Get the GPy.kern kernel by matching the given string kernel identifier.

        Parameters:
            kernel (str): Kernel string such as 'RBF' or depending on the surrogate also product and sum kernels
                such as 'RBF+Matern52'.

        Returns:
            GPy.kern: GPy kernel object. Currently, for sum and product kernels, the initial hyperparameters are the
            same for all kernels.
        """

        try:
            if not any(operator in kernel for operator in ('+', '*')):
                # Single kernel
                return getattr(self.GPy.kern, kernel)(self.ndim,
                                                      lengthscale=self.hyperparameters.get('length_scale', [1]),
                                                      variance=self.hyperparameters.get('sigma_f', 1)**2,
                                                      ARD=len(self.hyperparameters.get('length_scale', [1])) > 1)
            else:
                from re import split
                full_str = split('([+*])', kernel)
                kern = []
                for key in full_str:
                    kern += [key if key in ('+', '*') else
                               'self.GPy.kern.{}({}, lengthscale={}, variance={})'
                                   .format(key, self.ndim,
                                           self.hyperparameters.get('length_scale', [1]),
                                           self.hyperparameters.get('sigma_f', 1)**2)]
                return eval(''.join(kern))
        except AttributeError:
            raise RuntimeError("Kernel {} is not implemented.".format(kernel))

    def optimize(self, return_hess_inv=False, **opt_kwargs):
        """For hyperparameter optimization the GPy base optimization is used.

        Currently, the inverse Hessian can not be retrieved, which limits the active learning effectivity.

        Parameters:
            return_hess_inv (bool): Is not considered currently.
            opt_kwargs: Keyword arguments used directly in the GPy base optimization.
        """

        self.model.optimize(**opt_kwargs)
        self._set_hyperparameters_from_model()
        self.print_hyperparameters("Optimized")

    def _set_hyperparameters_from_model(self):
        r"""Helper function to set the hyperparameter dict from the model.

        It depends on whether it is a single kernel or a combined one.
        """
        if hasattr(self.model.kern, 'lengthscale'):
            self.hyperparameters['length_scale'] = self.model.kern.lengthscale.values
            self.hyperparameters['sigma_f'] = np.sqrt(self.model.kern.variance)
            self.hyperparameters['sigma_n'] = np.sqrt(self.model.likelihood.variance)
        elif hasattr(self.model.kern, 'parts'):
            for part in self.model.kern.parts:
                for key, value in zip(part.parameter_names(), part.param_array):
                    value = np.atleast_1d(value)
                    if key == 'lengthscale':
                        self.hyperparameters['length_scale'] = value
                    elif key == 'variance':
                        self.hyperparameters['sigma_f'] = np.sqrt(value)
                    else:
                        self.hyperparameters[key] = value
            noise_var = self.model.likelihood.gaussian_variance(
                self.model.Y_metadata).reshape(-1, 1, order='F')
            self.hyperparameters['sigma_n'] = np.sqrt(np.max(noise_var, axis=0))
        self.decode_hyperparameters()

    def special_hyperparameter_decoding(self, key, value):
        has_ard = any(p.ARD for p in self.model.kern.parts) if hasattr(self.model.kern, 'parts') else self.model.kern.ARD
        if key == 'length_scale' and self.ndim > 1 and not has_ard:
            return np.atleast_1d(np.linalg.norm(value))
        elif key in ('sigma_f', 'sigma_n') and len(value) > 1:
            return np.atleast_1d(np.max(value))
        return value


@Surrogate.register("CoregionalizedGPy")
class CoregionalizedGPySurrogate(GPySurrogate):

    def pre_train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
                  fixed_sigma_n=base_defaults['fixed_sigma_n']):
        super().pre_train(X, y, kernel, hyperparameters, fixed_sigma_n)
        self.output_ndim = self.ytrain.shape[-1]

    def train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
              fixed_sigma_n=base_defaults['fixed_sigma_n']):
        self.pre_train(X, y, kernel, hyperparameters, fixed_sigma_n)
        icm = self.GPy.util.multioutput.ICM(input_dim=self.ndim, num_outputs=self.output_ndim, kernel=self.kernel)
        _X = self.output_ndim * [self.Xtrain]
        _y = [self.ytrain[:, d].reshape(-1, 1) for d in range(self.output_ndim)]
        self.model = self.GPy.models.GPCoregionalizedRegression(_X, _y, kernel=icm)
        self.optimize()
        self.post_train()

    def add_training_data(self, X, y):
        self.Xtrain, self.ytrain = np.concatenate([self.Xtrain, X], axis=0), np.concatenate([self.ytrain, y], axis=0)
        self.encode_training_data()
        new_Xtrain = np.empty((self.Xtrain.shape[0] * self.output_ndim, self.Xtrain.shape[-1] + 1))
        for d in range(self.output_ndim):
            start_idx = self.Xtrain.shape[0] * d
            end_idx = start_idx + self.Xtrain.shape[0]
            new_Xtrain[start_idx:end_idx] = np.hstack([self.Xtrain, np.ones((self.Xtrain.shape[0], 1)) * d])
        new_noise_dict = {'output_index': new_Xtrain[:, -1:].astype(int)}
        self.model.Y_metadata = new_noise_dict
        self.model.set_XY(new_Xtrain, self.ytrain.reshape(-1, 1, order='F'))
        self.decode_training_data()

    def set_ytrain(self, y):
        self.ytrain = np.atleast_2d(y.copy())
        self.encode_training_data()
        self.model.set_Y(self.ytrain.reshape(-1, 1, order='F'))
        self.decode_training_data()

    def predict(self, Xpred, add_data_variance=True):
        Xpred = super().pre_predict(Xpred)
        ymean = np.empty((Xpred.shape[0], self.output_ndim))
        yvar = ymean.copy()
        for d in range(self.output_ndim):
            newX = np.hstack([Xpred, np.ones((Xpred.shape[0], 1)) * d])
            noise_dict = {'output_index': newX[:, -1:].astype(int)}
            ym, yv = self.model.predict(newX, Y_metadata=noise_dict)
            ymean[:, d], yvar[:, d] = ym.flatten(), yv.flatten()
        ymean, yvar = self.decode_predict_data(ymean, yvar)
        return ymean, yvar

    def save_model(self, path):
        # GPy does not support to_dict method for Coregionalization kernel yet.
        from profit.util.file_handler import FileHandler
        from os.path import splitext
        filepath, ending = splitext(path)
        if ending != '.pkl':
            print(f"Saving to '{ending}' not implemented yet for {self.__class__.__name__} surrogate."
                  " Saving to '.pkl' file instead.")
        save_list = [self.model, self.Xtrain, self.ytrain,
                     str([enc.repr for enc in self.input_encoders]),
                     str([enc.repr for enc in self.output_encoders])]
        FileHandler.save(filepath + '.pkl', save_list)

    @classmethod
    def load_model(cls, path):
        from profit.util.file_handler import FileHandler
        from profit.sur.encoders import Encoder
        from numpy import array  # needed for eval of arrays

        self = cls()
        try:
            self.model, self.Xtrain, self.ytrain, input_encoder_str, output_encoder_str = FileHandler.load(path, as_type='raw')
        except (OSError, FileNotFoundError):
            from os.path import splitext
            print("File not found. Try changing file ending to '.pkl'!")
            path = splitext(path)[0] + '.pkl'
            # Load multi-output model from pickle file
            self.model, self.Xtrain, self.ytrain, input_encoder_str, output_encoder_str = FileHandler.load(path, as_type='raw')

        for enc in eval(input_encoder_str):
            self.add_input_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))
        for enc in eval(output_encoder_str):
            self.add_output_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))
        self.output_ndim = int(max(self.model.X[:, -1])) + 1

        # Initialize the encoder by encoding and decoding the training data once.
        self.encode_training_data()
        self.decode_training_data()

        self.kernel = self.model.kern
        self.ndim = self.Xtrain.shape[-1]
        self._set_hyperparameters_from_model()
        self.trained = True
        self.print_hyperparameters("Loaded")
        return self

    def special_hyperparameter_decoding(self, key, value):
        if key == 'length_scale' and self.ndim > 1 and not self.model.kern.parts[0].ARD:
            return np.atleast_1d(np.linalg.norm(value))
        elif key in ('sigma_f', 'sigma_n') and len(value) > 1:
            return np.atleast_1d(np.max(value))
        elif key in ('W', 'kappa'):
                new_value = value
                for enc in self.output_encoders:
                    if enc.label == 'Normalization':
                        new_value = enc.decode_hyperparameters(value)
                return np.atleast_1d(np.min(new_value))
        return value



from profit.util.base_class import CustomABC
import numpy as np


class Encoder(CustomABC):
    r"""Base class to handle encoding and decoding of the input and output data before creating the surrogate model.

    The base class itself does nothing. It delegates the encoding process to the childs
    which are called by their registered labels.

    Parameters:
        columns (list[int]): Columns of the data the encoder acts on.
        parameters (dict): Miscellaneous parameters stored during encoding, which are needed for decoding. E.g. the
            scaling factor during normalization.

    Attributes:
        label (str): Label of the encoder class.
    """
    labels = {}

    def __init__(self, columns, parameters=None):
        self.label = self.__class__.get_label()
        self.parameters = {key: np.array(values) for key, values in parameters.items()} if parameters else {}
        self.columns = columns

    @property
    def repr(self):
        """Easy to handle representation of the encoder for saving and loading.
        Returns:
            list: List of all relevant information to reconstruct the encoder.
                (label, columns, parameters dict)
        """
        parameters_dict = {}
        for key, values in self.parameters.items():
            try:
                parameters_dict[key] = values.tolist()
            except AttributeError:
                parameters_dict[key] = [values]
        return {'class': self.label, 'columns': self.columns, 'parameters': parameters_dict}

    def encode(self, x):
        """Applies the encoding function on given columns.

        Parameters:
            x (ndarray): Array to which the encoding is applied.

        Returns:
            ndarray: An encoded copy of the array x.
        """
        _x = x.copy()
        _x[:, self.columns] = self.encode_func(_x[:, self.columns])
        return _x

    def decode(self, x):
        """Applies the decoding function on given columns.

        Parameters:
            x (ndarray): Array to which the decoding is applied.

        Returns:
            ndarray: A decoded copy of the array x.
        """
        _x = x.copy()
        _x[:, self.columns] = self.decode_func(_x[:, self.columns])
        return _x

    def decode_hyperparameters(self, value):
        """Decoder for the surrogate hyperparameters, as the direct model uses encoded values.
        As a default, the unchanged value is returned.

        Parameters:
            value (np.array): The (encoded) value of the hyperparameter.
        Returns:
            np.array: Decoded value.
        """
        return value

    def decode_variance(self, variance):
        return variance

    def encode_func(self, x):
        r"""
        Returns:
            ndarray: Function used for decoding the data. E.g. $\log_{10}(x)$.
        """
        return x

    def decode_func(self, x):
        r"""
        Returns:
            ndarray: Inverse transform of the encoding function. For an encoding of $\log_{10}(x)$ this
                would be $10^x$.
        """
        return x


@Encoder.register("Exclude")
class ExcludeEncoder(Encoder):
    """Excludes specific columns from the fit. Afterwards they are inserted at the same position.

    Variables:
        excluded_values (np.array): Slice of the input data which is excluded.
    """

    def encode(self, x):
        self.parameters['excluded_values'] = x[:, self.columns]
        return x[:, [i for i in range(x.shape[-1]) if i not in self.columns]]

    def decode(self, x):
        for idx, col in enumerate(self.columns):
            insert_x = self.parameters['excluded_values'][:, idx]
            x = np.insert(x, col, insert_x, axis=1)
        return x


@Encoder.register('Log10')
class Log10Encoder(Encoder):
    r"""Transforms the specified columns with $log_{10}$. This is done for LogUniform variables by default."""

    def encode_func(self, x):
        return np.log10(x)

    def decode_func(self, x):
        return 10**x


@Encoder.register('Normalization')
class Normalization(Encoder):
    r"""Normalization of the specified columns. Usually this is done for all input and output,
        so the surrogate can fit on a (0, 1)^n cube with zero mean and unit variance.

        $$
        \begin{align}
        x' &= (x - x_{min}) / (x_{max} - x_{min}) \\
        x & = (x_{max} - x_{min}) * x' + x_{min}
        \end{align}
        $$

    Parameters:
        xmax (np.array): Max. value of the data for each column.
        xmin (np.array): Min. value of the data for each column.
        xmean (np.array): Mean value of the data for each column.
        xstd (np.array): Standard deviation of the data for each column.
        xmax_centered (np.array): Max. value of the data after mean and variance standardization.
        xmin_centered (np.array): Min. value of the data after mean and variance standardization.
    """

    def encode(self, x):
        if self.parameters.get('xmean') is None:
            self.parameters['xmean'] = x[:, self.columns].mean(axis=0)
        if self.parameters.get('xstd') is None:
            self.parameters['xstd'] = np.maximum(x[:, self.columns].std(axis=0), 1e-10)
        if self.parameters.get('xmax') is None:
            self.parameters['xmax'] = x[:, self.columns].max(axis=0)
        self.parameters['xmax_centered'] = (self.parameters['xmax'] - self.parameters['xmean']) / \
                                           self.parameters['xstd']
        if self.parameters.get('xmin') is None:
            self.parameters['xmin'] = x[:, self.columns].min(axis=0)
        self.parameters['xmin_centered'] = (self.parameters['xmin'] - self.parameters['xmean']) / \
                                           self.parameters['xstd']
        return super().encode(x)

    def encode_func(self, x):
        _x = (x - self.parameters['xmean']) / self.parameters['xstd']
        _x = (_x - self.parameters['xmin_centered']) / \
             np.maximum(self.parameters['xmax_centered'] - self.parameters['xmin_centered'], 1e-10)
        return _x

    def decode_func(self, x):
        _x = x * (self.parameters['xmax_centered'] - self.parameters['xmin_centered']) + \
             self.parameters['xmin_centered']
        _x = _x * self.parameters['xstd'] + self.parameters['xmean']
        return _x

    def decode_hyperparameters(self, value):
        """Decode surrogate's hyperparameters. Distinguish between length_scale (only input_encoders)
        and sigma_f, sigma_n (only output_encoders) done in profit.sur.gp.gaussian_process.GaussianProcess.

        Parameters:
            value (np.array): The (encoded) value of the hyperparameter.
        Returns:
            np.array: Decoded value.
        """
        _value = value * (self.parameters['xmax_centered'] - self.parameters['xmin_centered'])
        _value = _value * self.parameters['xstd']
        return _value

    def decode_variance(self, variance):
        _variance = variance * (self.parameters['xmax_centered'] - self.parameters['xmin_centered']) ** 2
        _variance = _variance * self.parameters['xstd']**2
        return _variance


@Encoder.register("PCA")
class PCA(Encoder):

    def __init__(self, columns=(), parameters=None):
        if parameters is None:
            parameters = {}
        if 'tol' not in parameters:
            parameters['tol'] = 1e-2

        super().__init__(columns, parameters)
        if self.parameters.get('ytrain') is not None:
            self.init_eigvalues(self.parameters['ytrain'])
        else:
            self.ymean = None
            self.dy = None
            self.w = None
            self.Q = None

    def init_eigvalues(self, y):
        from scipy.linalg import eigh
        self.parameters['ytrain'] = y
        self.ymean = np.mean(y, 0)
        self.dy = y - self.ymean
        w, Q = eigh(self.dy.T @ self.dy)
        condi = w > self.parameters['tol']
        self.w = w[condi]
        self.Q = Q[:, condi]

    def encode(self, y):
        """
        Parameters:
            y: ntest sample vectors of length N.

        Returns:
            Expansion coefficients of y in eigenbasis.
        """
        if 'ytrain' not in self.parameters:
            self.init_eigvalues(y)
        return (y - self.ymean) @ self.Q

    def decode(self, z):
        """
        Parameters:
            z: Expansion coefficients of y in eigenbasis.

        Returns:
            Reconstructed ntest sample vectors of length N.
        """
        return self.ymean + (self.Q @ z.T).T

    @property
    def features(self):
        """
        Returns:
            neig feature vectors of length N.
        """
        return self.Q

    def decode_variance(self, variance):
        if variance.shape[-1] == self.Q.shape[-1]:
            # For multi-output data
            var_dec = np.empty((variance.shape[0], self.ymean.shape[0]))
            for row, var_enc in enumerate(variance):
                I = np.eye(var_enc.shape[0], var_enc.shape[0])
                var_dec[row, :] = np.diag(self.Q @ (var_enc * I @ self.Q.T))
            return var_dec
        return np.full_like(variance, np.nan)


@Encoder.register("KarhunenLoeve")
class KarhunenLoeve(PCA):

    def encode(self, y):
        """
        Parameters:
            y: ntest sample vectors of length N.

        Returns:
            Expansion coefficients of y in eigenbasis.
        """
        if 'ytrain' not in self.parameters:
            self.init_eigvalues(y)
        ntrain = self.dy.shape[0]
        ntest = y.shape[0]
        b = np.empty((ntrain, ntest))
        for i in range(ntrain):
            b[i,:] = (y - self.ymean) @ self.dy[i]
        yres = np.diag(1.0/self.w) @ self.Q.T @ b
        return yres.T

    def init_eigvalues(self, y):
        from scipy.linalg import eigh
        self.parameters['ytrain'] = y
        self.ymean = np.mean(y, 0)
        self.dy = y - self.ymean
        w, Q = eigh(self.dy @ self.dy.T)
        condi = w > self.parameters['tol']
        self.w = w[condi]
        self.Q = Q[:, condi]

    def decode(self, z):
        """
        Parameters:
            z: Expansion coefficients of y in eigenbasis.

        Returns:
            Reconstructed ntest sample vectors of length N.
        """
        return self.ymean + (self.dy.T @ (self.Q @ z.T)).T

    @property
    def features(self):
        return self.dy.T @ self.Q

    def decode_variance(self, variance):
        if variance.shape[-1] == self.Q.shape[-1]:
            # For multi-output data
            var_dec = np.empty((variance.shape[0], self.ymean.shape[0]))
            for row, var_enc in enumerate(variance):
                I = np.eye(var_enc.shape[0], var_enc.shape[0])
                var_dec[row, :] = np.diag((self.dy.T @ self.Q) @ (var_enc * I @ (self.dy.T @ self.Q).T))
            return var_dec
        return np.full_like(variance, np.nan)



import numpy as np
from profit.sur import Surrogate
from profit.sur.gp import GaussianProcess
from profit.defaults import fit_gaussian_process as defaults, fit as base_defaults


@Surrogate.register('Sklearn')
class SklearnGPSurrogate(GaussianProcess):
    """Surrogate for https://github.com/scikit-learn/scikit-learn Gaussian process.

    Attributes:
        model (sklearn.gaussian_process.GaussianProcessRegressor): Model object of Sklearn.
    """

    def __init__(self):
        super().__init__()
        self.model = None

    def train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
              fixed_sigma_n=base_defaults['fixed_sigma_n'], **kwargs):
        from sklearn.gaussian_process import GaussianProcessRegressor

        self.pre_train(X, y, kernel, hyperparameters, fixed_sigma_n)
        numeric_noise = self.hyperparameters['sigma_n'].item() ** 2 if self.fixed_sigma_n else 1e-5

        # Instantiate the model
        self.model = GaussianProcessRegressor(kernel=self.kernel, alpha=numeric_noise)

        # Train the model
        self.model.fit(self.Xtrain, self.ytrain)
        self.kernel = self.model.kernel_
        self.post_train()

    def post_train(self):
        self._set_hyperparameters_from_model()
        self.print_hyperparameters("Optimized")
        super().post_train()

    def add_training_data(self, X, y):
        """Add training points to existing data.

        Parameters:
            X (ndarray): Input points to add.
            y (ndarray): Observed output to add.
        """
        self.Xtrain = np.concatenate([self.Xtrain, X], axis=0)
        self.ytrain = np.concatenate([self.ytrain, y], axis=0)

    def set_ytrain(self, ydata):
        """Set the observed training outputs. This is important for active learning.

        Parameters:
            ydata (np.array): Full training output data.
        """
        self.ytrain = np.atleast_2d(ydata.copy())

    def predict(self, Xpred, add_data_variance=True):
        Xpred = self.pre_predict(Xpred)

        ymean, ystd = self.model.predict(Xpred, return_std=True)
        yvar = ystd.reshape(-1, 1)**2
        if add_data_variance:
            yvar = yvar + self.hyperparameters['sigma_n'] ** 2
        ymean, yvar = self.decode_predict_data(ymean, yvar)
        return ymean, yvar

    def save_model(self, path):
        """Save the SklGPSurrogate model to a pickle file. All attributes of the surrogate are loaded directly from the
        model.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """

        from pickle import dump
        dump(self.model, open(path, 'wb'))

    @classmethod
    def load_model(cls, path):
        """Load a saved SklGPSurrogate model from a pickle file and update its attributes.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            profit.sur.gaussian_process.SklearnGPSurrogate: Instantiated surrogate model.
        """

        from pickle import load

        self = cls()
        self.model = load(open(path, 'rb'))
        self.Xtrain = self.model.X_train_
        self.ytrain = self.model.y_train_
        self.kernel = self.model.kernel_
        self.ndim = self.Xtrain.shape[-1]
        self.fixed_sigma_n = self.model.alpha != 1e-5
        self.trained = True
        self._set_hyperparameters_from_model()
        self.print_hyperparameters("Loaded")
        return self

    def optimize(self, **opt_kwargs):
        """For hyperparameter optimization the Sklearn base optimization is used.

        Currently, the inverse Hessian can not be retrieved, which limits the active learning effectivity.

        Parameters:
            opt_kwargs: Keyword arguments used directly in the Sklearn base optimization.
        """
        self.encode_training_data()
        self.model.fit(self.Xtrain, self.ytrain, **opt_kwargs)
        self.decode_training_data()
        self._set_hyperparameters_from_model()
        self.print_hyperparameters("Optimized")

    def select_kernel(self, kernel):
        """Get the sklearn.gaussian_process.kernels kernel by matching the given kernel identifier.

        Parameters:
            kernel (str): Kernel string such as 'RBF' or depending on the surrogate also product and sum kernels
                such as 'RBF+Matern52'.

        Returns:
            sklearn.gaussian_process.kernels: Scikit-learn kernel object. Currently, for sum and product kernels,
            the initial hyperparameters are the same for all kernels.
        """

        from re import split
        from sklearn.gaussian_process import kernels as sklearn_kernels
        full_str = split('([+*])', kernel)
        try:
            kernel = []
            for key in full_str:
                kernel += [key if key in ('+', '*') else
                           getattr(sklearn_kernels, key)(length_scale=self.hyperparameters['length_scale'])]
        except AttributeError:
            raise RuntimeError("Kernel {} is not implemented.".format(kernel))

        if len(kernel) == 1:
            kernel = kernel[0]
        else:
            kernel = [str(key) if not isinstance(key, str) else key for key in kernel]
            kernel = eval(''.join(kernel))

        # Add scale and noise to kernel
        kernel *= sklearn_kernels.ConstantKernel(constant_value=1/self.hyperparameters['sigma_f'].item() ** 2)
        if not self.fixed_sigma_n:
            kernel += sklearn_kernels.WhiteKernel(noise_level=self.hyperparameters['sigma_n'].item() ** 2)

        return kernel

    def _set_hyperparameters_from_model(self):
        r"""Helper function to set the hyperparameter dict from the model.

        It depends on whether $\sigma_n$ is fixed.
        Currently this is only stable for single kernels and not for Sum and Prod kernels.
        """
        if self.fixed_sigma_n:
            self.hyperparameters['length_scale'] = np.atleast_1d(self.model.kernel_.k1.length_scale)
            self.hyperparameters['sigma_f'] = np.sqrt(np.atleast_1d(1 / self.model.kernel_.k2.constant_value))
            self.hyperparameters['sigma_n'] = np.sqrt(np.atleast_1d(self.model.alpha))
        else:
            self.hyperparameters['length_scale'] = np.atleast_1d(self.model.kernel_.k1.k1.length_scale)
            self.hyperparameters['sigma_f'] = np.sqrt(np.atleast_1d(1 / self.model.kernel_.k1.k2.constant_value))
            self.hyperparameters['sigma_n'] = np.sqrt(np.atleast_1d(self.model.kernel_.k2.noise_level))
        self.decode_hyperparameters()


# Draft for Scikit-learn implementation of LinearEmbedding kernel of Garnett (2014)
from sklearn.gaussian_process.kernels import Kernel, Hyperparameter, StationaryKernelMixin, NormalizedKernelMixin
class LinearEmbedding(StationaryKernelMixin, NormalizedKernelMixin, Kernel):

    def __init__(self, dims, length_scale=np.array([1.0]), length_scale_bounds=(1e-5, 1e5)):
        self.length_scale = length_scale
        self.dims = dims
        self.length_scale_bounds = length_scale_bounds

    @property
    def hyperparameter_length_scale(self):
        return Hyperparameter("length_scale", "numeric",
                              self.length_scale_bounds,
                              len(self.length_scale))

    def __call__(self, X, Y=None, eval_gradient=False):
        X = np.atleast_2d(X)
        R = self.length_scale.reshape(self.dims)
        X1 = X @ R.T
        X2 = X @ R.T if Y is None else Y @ R.T
        dX = X1[:, np.newaxis, :] - X2[np.newaxis, :, :]
        dX_sq = np.linalg.norm(dX, axis=-1) ** 2
        K = np.exp(-0.5 * dX_sq)
        if eval_gradient:
            K_gradient = np.einsum('ijk,kl', dX ** 2, R) * K[..., np.newaxis]
            return K, K_gradient
        return K

    def __repr__(self):
        return "{0}(length_scale=[{1}])".format(self.__class__.__name__, ", ".join(map("{0:.3g}".format,
                                                                                       self.length_scale)))



import numpy as np

from profit.al import ActiveLearning
from profit.defaults import active_learning as base_defaults
from profit.defaults import al_algorithm_simple as defaults


@ActiveLearning.register("simple")
class SimpleAL(ActiveLearning):
    """Simple active learning algorithm based on a surrogate model and an acquisition function to find next candidates.

    Parameters:
        surrogate (profit.sur.Surrogate): Surrogate used for fitting.
        acquisition_function (str/profit.al.acquisition_functions.AcquisitionFunction): Acquisition function used for
            selecting the next candidates.
    Attributes:
        search_space (dict[str, np.array]): np.linspace for each AL input variable.
        Xpred (np.array): Matrix of the candidate points built with np.meshgrid.
    """
    labels = {}

    def __init__(self, runner, variables, surrogate, ntrain, nwarmup=base_defaults['nwarmup'],
                 batch_size=base_defaults['batch_size'], acquisition_function=defaults['acquisition_function'],
                 convergence_criterion=base_defaults['convergence_criterion'], nsearch=base_defaults['nsearch'],
                 make_plot=base_defaults['make_plot']):
        from profit.al.aquisition_functions import AcquisitionFunction

        super().__init__(runner, variables, ntrain, nwarmup, batch_size, convergence_criterion, nsearch, make_plot)
        self.surrogate = surrogate

        self.search_space = {var.name: np.linspace(*var.constraints, nsearch)
                             for var in variables.list if var.kind.lower() in 'activelearning'}

        Xpred = [var.create_Xpred((nsearch, 1)) for var in variables.input_list]
        self.Xpred = np.hstack([xi.flatten().reshape(-1, 1) for xi in np.meshgrid(*Xpred)])

        if issubclass(acquisition_function.__class__, AcquisitionFunction):
            self.acquisition_function = acquisition_function
        elif isinstance(acquisition_function, dict):
            label = acquisition_function['class']
            params = {key: value for key, value in acquisition_function.items() if key != 'class'}
            self.acquisition_function = AcquisitionFunction[label](self.Xpred, self.surrogate, self.variables, **params)
        else:
            self.acquisition_function = AcquisitionFunction[acquisition_function](self.Xpred, self.surrogate,
                                                                                  self.variables)
        # Set variable parameters for acquisiton_function
        for p in self.acquisition_function.al_parameters:
            self.acquisition_function.al_parameters[p] = getattr(self, p)

    def warmup(self, save_intermediate=base_defaults['save_intermediate']):
        """To get data for active learning, sample initial points randomly."""
        from profit.util.variable import halton
        params_array = [{} for _ in range(self.nwarmup)]
        halton_seq = halton(size=(self.nwarmup, self.variables.input.shape[-1]))
        for idx, values in enumerate(self.variables.named_input[:self.nwarmup]):
            names = values.dtype.names
            for col, key in enumerate(names):
                if key in self.search_space:
                    minv = self.search_space[key][0]
                    maxv = self.search_space[key][-1]
                    rand = minv + (maxv - minv) * halton_seq[idx, col]
                else:
                    rand = values[key][0]
                params_array[idx][key] = rand

        self.runner.spawn_array(params_array, blocking=True)
        self.update_data()

        self.surrogate.train(self.variables.input[:self.nwarmup], self.variables.output[:self.nwarmup])

        if save_intermediate:
            self.save_intermediate(**save_intermediate)

        if self.make_plot:
            self.plot()

    def learn(self, resume_from=base_defaults['resume_from'], save_intermediate=base_defaults['save_intermediate']):
        from time import time
        from tqdm import tqdm

        st = time()
        kstart = resume_from if resume_from is not None else self.nwarmup
        for krun in tqdm(range(kstart, self.ntrain, self.batch_size)):
            """
            1. find next candidates (and assign input)
            2. update runs
            3. assign output (if mcmc is accepted, else delete input)
            4. optimize surrogate
            """
            self.krun = krun

            # Set variable parameters inside the acquisition function
            al_params = {key: getattr(self, key) for key in self.acquisition_function.al_parameters}
            self.acquisition_function.set_al_parameters(**al_params)

            candidates = self.find_next_candidates()
            self.update_run(candidates)
            self.surrogate.set_ytrain(self.variables.output[:krun + self.batch_size])
            self.surrogate.optimize()

            if save_intermediate:
                self.save_intermediate(**save_intermediate)

            if self.make_plot:
                self.plot()

        print("Runtime main loop: {}".format(time() - st))
        if self.make_plot:
            from matplotlib.pyplot import show
            show()

    def find_next_candidates(self):
        """Find the next candidates using the acquisition function's method find_next_candidates.

        Returns:
            np.array: Next training points.
        """

        candidates = self.acquisition_function.find_next_candidates(self.batch_size)
        print("\nNext candidates: {}".format(candidates))
        return candidates

    def update_run(self, candidates):
        super().update_run(candidates)
        self.update_data()

    def save(self, path):
        self.surrogate.save_model(path)

    def plot(self):
        """Plot the progress of the AL learning."""
        from matplotlib.pyplot import figure, scatter
        figure()
        self.surrogate.plot(self.Xpred)
        # scatter(self.variables.input[self.krun:self.krun + self.batch_size],
        #        self.variables.output[self.krun:self.krun + self.batch_size],
        #        marker='x', c='r')

    @classmethod
    def from_config(cls, runner, variables, config, base_config):
        from profit.sur import Surrogate
        surrogate = Surrogate.from_config(base_config['fit'], base_config)
        return cls(runner, variables, surrogate, ntrain=base_config['ntrain'], nwarmup=config['nwarmup'],
                   batch_size=config['batch_size'], acquisition_function=config['algorithm']['acquisition_function'],
                   convergence_criterion=config['convergence_criterion'], nsearch=config['nsearch'],
                   make_plot=base_config['ui']['plot'])



import numpy as np
from profit.sur import Surrogate
from profit.sur.gp import GaussianProcess
from profit.defaults import fit_gaussian_process as defaults, fit as base_defaults


@Surrogate.register('Custom')
class GPSurrogate(GaussianProcess):
    """Custom GP model made from scratch.

     Supports custom Python and Fortran kernels with analytic derivatives and advanced Active Learning.

    Attributes:
        hess_inv (ndarray): Inverse Hessian matrix which is required for active learning.
            It is calculated during the hyperparameter optimization.
    """

    from .backend import gp_functions

    def __init__(self):
        super().__init__()
        self.hess_inv = None

    @property
    def Ky(self):
        """Full training covariance matrix as defined in the kernel
        including data noise as specified in the hyperparameters.
        """
        return self.kernel(self.Xtrain, self.Xtrain, **self.hyperparameters)

    @property
    def alpha(self):
        r"""Convenient matrix-vector product of the inverse training matrix and the training output data.
        The equation is solved either exactly or with a least squares approximation.

        $$
        \begin{equation}
        \alpha = K_y^{-1} y_{train}
        \end{equation}
        $$
        """
        try:
            return np.linalg.solve(self.Ky, self.ytrain)
        except np.linalg.LinAlgError:
            return np.linalg.lstsq(self.Ky, self.ytrain, rcond=1e-15)[0]

    def train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
              fixed_sigma_n=base_defaults['fixed_sigma_n'], eval_gradient=True, return_hess_inv=False):
        """After initializing the model with a kernel function and initial hyperparameters,
        it can be trained on input data X and observed output data y by optimizing the model's hyperparameters.
        This is done by minimizing the negative log likelihood.

        Parameters:
            X (ndarray): (n, D) array of input training data.
            y (ndarray): (n, 1) array of training output.
                Currently, only scalar output is supported.
            kernel (str/object): Identifier of kernel like 'RBF' or directly the kernel object of the
                specific surrogate.
            hyperparameters (dict): Hyperparameters such as length_scale, variance and noise.
                Taken either from given parameter, config file or inferred from the training data.
                The hyperparameters can be different depending on the kernel. E.g. The length_scale can be a scalar,
                a vector of the size of the training data, or for the custom LinearEmbedding kernel a matrix.
            fixed_sigma_n (bool): Indicates if the data noise should be optimized or not.
            eval_gradient (bool): Whether the gradients of the kernel and negative log likelihood are
                explicitly used in the scipy optimization or numerically calculated inside scipy.
            return_hess_inv (bool): Whether to the attribute hess_inv after optimization. This is important
                for active learning.
        """
        self.pre_train(X, y, kernel, hyperparameters, fixed_sigma_n)
        if self.input_encoders or self.output_encoders:
            print("For now, encoding is not supported for this surrogate. The model is created without encoding.")
            self.decode_training_data()

        # Find best hyperparameters
        self.optimize(fixed_sigma_n=self.fixed_sigma_n, eval_gradient=eval_gradient, return_hess_inv=return_hess_inv)
        self.post_train()

    def post_train(self):
        self.trained = True

    def predict(self, Xpred, add_data_variance=True):
        Xpred = super().pre_predict(Xpred)
        # Encoding is not supported yet for this surrogate.
        for enc in self.input_encoders[::-1]:
            Xpred = enc.decode(Xpred)

        # Skip data noise sigma_n in hyperparameters
        prediction_hyperparameters = {key: value for key, value in self.hyperparameters.items() if key != 'sigma_n'}

        # Calculate conditional mean and covariance functions
        Kstar = self.kernel(self.Xtrain, Xpred, **prediction_hyperparameters)
        Kstarstar_diag = np.diag(self.kernel(Xpred, Xpred, **prediction_hyperparameters))
        fstar = Kstar.T @ self.alpha
        vstar = Kstarstar_diag - np.diag((Kstar.T @ (self.gp_functions.invert(self.Ky) @ Kstar)))
        vstar = np.maximum(vstar, 1e-10)  # Assure a positive variance
        if add_data_variance:
            vstar = vstar + self.hyperparameters['sigma_n']**2
        return fstar, vstar.reshape(-1, 1)  # Return predictive mean and variance

    def add_training_data(self, X, y):
        """Add training points to existing data. This is important for active learning.

        Only the training dataset is updated, but the hyperparameters are not optimized yet.

        Parameters:
            X (ndarray): Input points to add.
            y (ndarray): Observed output to add.
        """

        # TODO: Update Ky by applying the Sherman-Morrison-Woodbury formula?
        self.Xtrain = np.concatenate([self.Xtrain, X], axis=0)
        self.ytrain = np.concatenate([self.ytrain, y], axis=0)

    def set_ytrain(self, ydata):
        """Set the observed training outputs. This is important for active learning.

        Parameters:
            ydata (np.array): Full training output data.
        """
        self.ytrain = np.atleast_2d(ydata.copy())

    def get_marginal_variance(self, Xpred):
        from .backend.gp_functions import marginal_variance_BBQ
        self.optimize(fixed_sigma_n=self.fixed_sigma_n, eval_gradient=True,
                      return_hess_inv=True)
        assert self.hess_inv is not None

        _, fvar = self.predict(Xpred)
        return marginal_variance_BBQ(self.Xtrain, self.ytrain, Xpred, self.kernel,
                                     self.hyperparameters, self.hess_inv, self.fixed_sigma_n,
                                     alpha=self.alpha, predictive_variance=fvar)

    def save_model(self, path):
        """Saves the model as dict to a .hdf5 file.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """
        from profit.util.file_handler import FileHandler

        saved_attrs = ('trained', 'fixed_sigma_n', 'Xtrain', 'ytrain', 'ndim', 'kernel', 'hyperparameters')
        save_dict = {attr: getattr(self, attr) for attr in saved_attrs}

        # Convert the kernel class object to a string, to be able to save it in the .hdf5 file
        if not isinstance(save_dict['kernel'], str):
            save_dict['kernel'] = self.kernel.__name__
        FileHandler.save(path, save_dict)

    @classmethod
    def load_model(cls, path):
        """Loads a saved model from a .hdf5 file and updates its attributes.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            profit.sur.gaussian_process.GPSurrogate: Instantiated surrogate model.
        """

        from profit.util.file_handler import FileHandler

        sur_dict = FileHandler.load(path, as_type='dict')
        self = cls()

        for attr, value in sur_dict.items():
            setattr(self, attr, value)

        # Convert the kernel string back to the class object
        self.kernel = self.select_kernel(self.kernel)
        self.print_hyperparameters("Loaded")
        return self

    def select_kernel(self, kernel):
        """Convert the name of the kernel as string to the kernel class object of the surrogate.
        First search the kernels implemented in python, then the Fortran kernels.

        Parameters:
            kernel (str): Kernel string such as 'RBF'. Only single kernels are supported currently.

        Returns:
            object: Kernel object of the class. This is the function which builds the kernel and not
            the calculated covariance matrix.
        """

        # TODO: Rewrite fortran kernels and rename them to be explicit, like 'fRBF'
        try:
            from .backend import kernels as fortran_kernels
        except:
            pass
        from .backend import python_kernels

        try:
            return getattr(python_kernels, kernel)
        except AttributeError:
            try:
                return getattr(fortran_kernels, kernel)
            except AttributeError:
                raise RuntimeError(f"Kernel {kernel} not implemented.")

    def optimize(self, fixed_sigma_n=base_defaults['fixed_sigma_n'], eval_gradient=False, return_hess_inv=False):
        r"""Optimize the hyperparameters length_scale $l$, scale $\sigma_f$ and noise $\sigma_n$.

        As a backend, the scipy minimize optimizer is used.

        Parameters:
            fixed_sigma_n (bool): Indication if the data noise should also be optimized or not.
            eval_gradient (bool): Flag if the gradients of the kernel and negative log likelihood should be
                used explicitly or numerically calculated inside the optimizer.
            return_hess_inv (bool): Whether to set the inverse Hessian attribute hess_inv which is used to calculate the
                marginal variance in active learning.
        """
        ordered_hyp_keys = ('length_scale', 'sigma_f', 'sigma_n')
        a0 = np.concatenate([self.hyperparameters[key] for key in ordered_hyp_keys])
        opt_hyperparameters = self.gp_functions.optimize(self.Xtrain, self.ytrain, a0, self.kernel,
                                                         fixed_sigma_n=self.fixed_sigma_n or fixed_sigma_n,
                                                         eval_gradient=eval_gradient, return_hess_inv=return_hess_inv)
        if return_hess_inv:
            self.hess_inv = opt_hyperparameters[1]
            opt_hyperparameters = opt_hyperparameters[0]
        self._set_hyperparameters_from_model(opt_hyperparameters)
        self.print_hyperparameters("Optimized")

    def _set_hyperparameters_from_model(self, model_hyperparameters):
        # Set optimized hyperparameters
        last_idx = -1 if self.fixed_sigma_n else -2
        self.hyperparameters['length_scale'] = np.atleast_1d(model_hyperparameters[:last_idx])
        self.hyperparameters['sigma_f'] = np.atleast_1d(model_hyperparameters[last_idx])
        if not self.fixed_sigma_n:
            self.hyperparameters['sigma_n'] = np.atleast_1d(model_hyperparameters[-1])


@Surrogate.register("CustomMultiOutputGP")
class MultiOutputGPSurrogate(GaussianProcess):

    def __init__(self, child=GPSurrogate):
        super().__init__()
        self.child = child
        self.models = []

    def train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
              fixed_sigma_n=base_defaults['fixed_sigma_n'], return_hess_inv=False):
        self.pre_train(X, y)
        self.models = [self.child() for _ in range(self.output_ndim)]
        for dim, m in enumerate(self.models):
            m.train(self.Xtrain, self.ytrain[:, [dim]], self.kernel, self.hyperparameters, self.fixed_sigma_n,
                    False, return_hess_inv)
        self._set_hyperparameters_from_model()
        self.post_train()

    def _set_hyperparameters_from_model(self):
        for m in self.models:
            for k, v in m.hyperparameters.items():
                self.hyperparameters[k] = np.concatenate([self.hyperparameters[k], v])
        self.hyperparameters['length_scale'] = np.atleast_1d(np.linalg.norm(self.hyperparameters['length_scale']))
        self.hyperparameters['sigma_f'] = np.atleast_1d(np.max(self.hyperparameters['sigma_f']))
        self.hyperparameters['sigma_n'] = np.atleast_1d(np.max(self.hyperparameters['sigma_n']))
        self.decode_hyperparameters()

    def predict(self, Xpred, add_data_variance=True):
        Xpred = self.pre_predict(Xpred)
        ypred = np.empty((Xpred.shape[0], self.output_ndim))
        yvar = np.empty_like(ypred)

        for dim, m in enumerate(self.models):
            ypred[:, [dim]], yvar[:, [dim]] = m.predict(Xpred, add_data_variance)
        ypred, yvar = self.decode_predict_data(ypred, yvar)
        return ypred, yvar

    def add_training_data(self, X, y):
        start = self.Xtrain.shape[0]
        self.Xtrain, self.ytrain = np.concatenate([self.Xtrain, X], axis=0), np.concatenate([self.ytrain, y], axis=0)
        self.encode_training_data()
        for dim, m in enumerate(self.models):
            m.add_training_data(self.Xtrain[start:], self.ytrain[start:, [dim]])
        self.decode_training_data()

    def set_ytrain(self, y):
        for dim, m in enumerate(self.models):
            m.set_ytrain(y[:, [dim]])

        self.ytrain = np.atleast_2d(y.copy())

    def save_model(self, path):
        """Saves the model as dict to a .hdf5 file.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """

        from profit.util.file_handler import FileHandler
        save_dict = {attr: getattr(self, attr) for attr in ('Xtrain', 'ytrain', 'trained', 'output_ndim', 'hyperparameters')}
        save_dict['input_encoders'] = str([enc.repr for enc in self.input_encoders])
        save_dict['output_encoders'] = str([enc.repr for enc in self.output_encoders])
        for i, m in enumerate(self.models):
            save_dict[i] = {attr: getattr(m, attr)
                            for attr in
                            ('trained', 'fixed_sigma_n', 'Xtrain', 'ytrain',
                             'ndim', 'output_ndim', 'kernel', 'hyperparameters')}
            # Convert the kernel class object to a string, to be able to save it in the .hdf5 file
            if not isinstance(save_dict[i]['kernel'], str):
                save_dict[i]['kernel'] = m.kernel.__name__
        FileHandler.save(path, save_dict)

    @classmethod
    def load_model(cls, path):
        from profit.util.file_handler import FileHandler
        from profit.sur.encoders import Encoder
        from numpy import array  # needed for eval of arrays

        load_dict = FileHandler.load(path, as_type='dict')
        self = cls()
        self.trained = load_dict['trained']
        self.output_ndim = load_dict['output_ndim']
        self.Xtrain, self.ytrain = load_dict['Xtrain'], load_dict['ytrain']
        self.hyperparameters = load_dict['hyperparameters']

        for enc in eval(load_dict['input_encoders']):
            self.add_input_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))
        for enc in eval(load_dict['output_encoders']):
            self.add_output_encoder(Encoder[enc['class']](enc['columns'], enc['parameters']))

        # Initialize the encoder by encoding and decoding the training data once.
        self.encode_training_data()
        self.decode_training_data()

        self.models = [self.child() for _ in range(self.output_ndim)]

        for i, m in enumerate(self.models):
            for attr, value in load_dict[str(i)].items():
                setattr(m, attr, value)
            # Convert the kernel string back to the class object
            m.kernel = m.select_kernel(m.kernel)
            m.print_hyperparameters("Loaded")
        return self

    def optimize(self, **opt_kwargs):
        for m in self.models:
            m.optimize(**opt_kwargs)

    def special_hyperparameter_decoding(self, key, value):
        if len(value) > 1:
            return np.atleast_1d(np.linalg.norm(value)) if key == 'length_scale' else np.atleast_1d(np.max(value))
        return value

    def get_marginal_variance(self, Xpred):
        Xpred = self.encode_predict_data(Xpred)
        v = np.zeros((Xpred.shape[0], 1))
        for m in self.models:
            v += m.get_marginal_variance(Xpred)
        return v



"""proFit main script.

This script is called when running the `profit` command inside a shell.
"""

import sys
from argparse import ArgumentParser
from platform import python_version

from profit.config import BaseConfig
from profit.util import safe_path
from profit.util.variable import VariableGroup, Variable
from profit.defaults import base_dir as default_base_dir, config_file as default_config_file
from profit.run import Runner


def main():
    """Main command line interface"""
    from profit import __version__  # delayed to prevent cyclic import

    # Get parameters from shell input
    # display help by default: https://stackoverflow.com/questions/4042452/display-help-message-with-python-argparse-when-script-is-called-without-any-argu
    class MyParser(ArgumentParser):
        def error(self, message):
            sys.stderr.write(f"error: {message}\n")
            self.print_help(sys.stderr)
            self.exit(2)

    parser = MyParser(description=f"Probabilistic Response Model Fitting with Interactive Tools v{__version__}")
    subparsers = parser.add_subparsers(metavar="mode", dest="mode", required=True)
    subparsers.add_parser("run", help="start simulation runs")
    subparsers.add_parser("fit", help="fit data (e.g. with a Gaussian Process)")
    subparsers.add_parser("ui", help="interactively visualize results using dash")
    subparsers.choices["ui"].add_argument("--debug", action="store_true", help="activate debug mode for the Dash app")
    subparsers.add_parser("clean", help="remove temporary files, run directories and logs")
    subparsers.choices["clean"].add_argument("--all", action="store_true", help="remove input, output and model files")
    subparsers.add_parser("version", help="show version information")

    parser.add_argument('base_dir',
                        metavar='base-dir',
                        help='path to config file or directory containing profit.yaml (default: current working directory)',
                        default=default_base_dir, nargs='?')

    args, reminder = parser.parse_known_args()

    # parse additional arguments which weren't used by the subparser by a modified main parser again
    parser.usage = parser.format_usage()
    del parser._actions[1]  # without subparsers
    args = parser.parse_args(reminder, namespace=args)
    
    # `profit version` does not require a config
    if args.mode == "version":
        print(f"proFit {__version__}")
        import profit.sur.gp.backend
        try:
            from profit.sur.gp.backend import gpfunc
            print('with fortran backend')
        except ImportError:
            print('without fortran backend')
        print(f"python {python_version()}")
        return

    # Instantiate Config from the given file
    config_file = safe_path(args.base_dir, default=default_config_file)
    config = BaseConfig.from_file(config_file)
    variables = config.variable_group

    sys.path.append(config['base_dir'])

    # Select mode
    if args.mode == 'run':
        from tqdm import tqdm
        from profit.util import check_ndim
        from profit.util.file_handler import FileHandler

        runner = Runner.from_config(config['run'], config)  # Instantiate the runner
        if config['active_learning']['resume_from'] is not None:
            X = FileHandler.load(config['files']['input'])
            y = FileHandler.load(config['files']['output'])
            for v in variables.list:
                if v.name in X.dtype.names:
                    v.value[:X.shape[0]] = X[v.name]
                else:
                    v.value[:y.shape[0]] = y[v.name]
        FileHandler.save(config['files']['input'], variables.named_input)  # Save variables to input file

        # Check if active learning needs to be done instead of executing a normal run
        if 'activelearning' in (v.kind.lower() for v in variables.list):
            from profit.al.active_learning import ActiveLearning
            from profit.sur.sur import Surrogate

            runner.fill(variables.named_input)  # Prepare runner with input variables
            al = ActiveLearning.from_config(runner, variables, config['active_learning'], config)  # Instantiate the active learning algorithm
            try:
                save_intermediate = {'model_path': config['fit']['save'] or config['fit']['load'],
                                     'input_path': config['files']['input'],
                                     'output_path': config['files']['output']}
                if config['active_learning']['resume_from']:
                    resume_from = config['active_learning']['resume_from']
                    runner.fill_output(variables.named_output)
                    runner.next_run_id = resume_from
                    al.learn(resume_from=resume_from, save_intermediate=save_intermediate)
                else:
                    al.warmup(save_intermediate=save_intermediate)  # Execute warmup cycles
                    al.learn(save_intermediate=save_intermediate)  # Execute main learning loop
                FileHandler.save(config['files']['input'], variables.named_input)  # Save learned input variables
            finally:
                runner.cancel_all()  # Close all run processes
            if config['active_learning']['algorithm']['save']:
                al.save(config['active_learning']['algorithm']['save'])  # Save AL model
        else:
            # Normal (parallel) run
            params_array = [row[0] for row in variables.named_input]  # Structured array of input values
            try:
                runner.spawn_array(tqdm(params_array), blocking=True)  # Start runs
            except KeyboardInterrupt:
                runner.logger.info("Keyboard Interrupt")
            finally:
                runner.cancel_all()  # Close all run processes

                # Write runner output data into variables
                for key in runner.output_data.dtype.names:
                    variables[key].value = check_ndim(runner.output_data[key])

        if config['run']['clean']:
            runner.clean()
        if len(runner.failed):
            runner.logger.warning(f"{len(runner.failed)} runs failed")

        # Format output data for txt file and save
        formatted_output_data = variables.formatted_output \
            if config['files']['output'].endswith('.txt') else variables.named_output
        FileHandler.save(config['files']['output'], formatted_output_data)

    elif args.mode == 'fit':
        from numpy import arange, hstack, meshgrid
        from profit.util.file_handler import FileHandler
        from profit.sur.sur import Surrogate

        sur = Surrogate.from_config(config['fit'], config)  # Instantiate surrogate model

        # Train model if not yet trained
        if not sur.trained:
            x = FileHandler.load(config['files']['input'])
            y = FileHandler.load(config['files']['output'])
            x = hstack([x[key] for key in x.dtype.names])  # Convert structured to normal array
            y = hstack([y[key] for key in y.dtype.names])

            sur.train(x, y)

        if config['fit']['save']:
            sur.save_model(config['fit']['save'])  # Save surrogate model
        if config['ui']['plot']:
            # Make a simple plot of data and surrogate model
            # TODO: Rename to 'simple_plot' and introduce more options
            try:
                # Get prediction range from input or infer from training data
                xpred = [arange(minv, maxv, step) for minv, maxv, step in config['ui']['plot'].get('Xpred')]
                xpred = hstack([xi.flatten().reshape(-1, 1) for xi in meshgrid(*xpred)])
            except AttributeError:
                xpred = None
            sur.plot(xpred, independent=config['independent'], show=True)

    elif args.mode == 'ui':
        from profit.ui import init_app
        app = init_app(config)
        app.run_server(debug=args.debug)  # Start Dash server on localhost

    elif args.mode == 'clean':
        from shutil import rmtree
        from os import path, remove
        run_dir = config['run_dir']

        # Remove single run directories
        for krun in range(config['ntrain']):
            single_run_dir = path.join(run_dir, f'run_{krun:03d}')
            if path.exists(single_run_dir):
                rmtree(single_run_dir)

        if args.all:
            # Remove input and output files
            if path.exists(config['files']['input']):
                remove(config['files']['input'])
            if path.exists(config['files']['output']):
                remove(config['files']['output'])
            if path.exists(config['fit']['save']):
                remove(config['fit']['save'])

        # Cleanup runner
        runner = Runner.from_config(config['run'], config)
        runner.clean()
        try:
            # Remove log
            rmtree(config['run']['log_path'])
        except FileNotFoundError:
            pass
        if path.exists('runner.log'):
            remove('runner.log')


if __name__ == '__main__':
    main()



"""
Testcases for mockup simulations:
- all input variables are Halton sequences to ensure reproducibility
- 1D function
    - .hdf5 input/output
    - multi output
- 2D function (Rosenbrock)
    - .txt input/output
- 2D function (Fermi)
    - one Halton (Temperature) and one independent (Energy) variable (simulation returns a vector)
    - .json template
    - .hdf5 input/output
    - .hdf5 output of single runs with custom interface
"""

from profit.config import BaseConfig
from profit.sur import Surrogate
from profit.util.file_handler import FileHandler
from os import path, remove, chdir, getcwd
from subprocess import run
from numpy import array, allclose
from shutil import rmtree
from pytest import fixture


@fixture(autouse=True)
def chdir_pytest():
    pytest_root_dir = getcwd()
    chdir(path.dirname(path.abspath(__file__)))
    yield
    chdir(pytest_root_dir)


# Allow a large range of parameters, just ensure that it is approximately at the same scale
NLL_ATOL = 1e3
PARAM_RTOL = 2
TIMEOUT = 30  # seconds


def clean(config):
    """Delete run directories and input/outpt files after the test."""

    for krun in range(config.get('ntrain')):
        single_dir = path.join(config.get('run_dir'), f'run_{krun:03d}')
        if path.exists(single_dir):
            rmtree(single_dir)
    if path.exists(config['run']['interface'].get('path')):
        remove(config['run']['interface'].get('path'))
    if path.exists(config['files'].get('input')):
        remove(config['files'].get('input'))
    if path.exists(config['files'].get('output')):
        remove(config['files'].get('output'))
    if path.exists(config['run'].get('log_path')):
        rmtree(config['run'].get('log_path'))


def test_1D():
    """Test a simple function f(u) = cos(10*u) + u."""

    config_file = 'study_1D/profit_1D.yaml'
    config = BaseConfig.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'GPy'
        assert sur.trained
        assert sur.model.kern.name == 'rbf'
        assert allclose(sur.hyperparameters['length_scale'], 0.23521412, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_f'], 1.56475873, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_n'], 5.26616713e-05, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)


def multi_test_1d(study, config_file, output_file):
    """ test 1D with different config files """
    config_file = path.join(study, config_file)
    output_file = path.join(study, output_file)
    config = BaseConfig.from_file(config_file)
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        output = FileHandler.load(output_file)
        assert output.shape == (7, 1)
        assert all(output['f'] - array([0.7836, -0.5511, 1.0966, 0.4403, 1.6244, -0.4455, 0.0941]).reshape((7, 1))
                   < 1e-4)
    finally:
        clean(config)


def test_custom_post1():
    """ test 1D with custom postprocessor """
    multi_test_1d('./study_custom_post1', 'profit_custom_post1.yaml', 'output_custom_post1.hdf5')


def test_custom_post2():
    """ test 1D with custom postprocessor (wrap syntax) """
    multi_test_1d('./study_custom_post2', 'profit_custom_post2.yaml', 'output_custom_post2.hdf5')


def test_custom_worker1():
    """ test 1D with custom worker (custom entry point) """
    multi_test_1d('./study_custom_worker1', 'profit_custom_worker1.yaml', 'output_custom_worker1.hdf5')


def test_custom_worker2():
    """ test 1D with custom worker (integrated, custom main) """
    multi_test_1d('./study_custom_worker2', 'profit_custom_worker2.yaml', 'output_custom_worker2.hdf5')


def test_custom_worker3():
    """ test 1D with custom worker (integrated, custom run) """
    multi_test_1d('./study_custom_worker3', 'profit_custom_worker3.yaml', 'output_custom_worker3.hdf5')


def test_custom_worker4():
    """ test 1D with custom worker (integrated, custom run) """
    multi_test_1d('./study_custom_worker4', 'profit_custom_worker4.yaml', 'output_custom_worker4.hdf5')


def test_multi_output():
    """Test a 1D function with two outputs."""
    config_file = 'study_multi_output/profit_multi_output.yaml'
    config = BaseConfig.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'CustomMultiOutputGP'
        assert sur.trained
        length_scale = [0.22920459, 0.204117]
        sigma_f = [1.38501625, 1.20300593]
        sigma_n = [3.37481098e-05, 8.76414971e-05]
        for i, m in enumerate(sur.models):
            assert m.kernel.__name__ == 'RBF'
            assert allclose(m.hyperparameters['length_scale'], length_scale[i], rtol=PARAM_RTOL)
            assert allclose(m.hyperparameters['sigma_f'], sigma_f[i], rtol=PARAM_RTOL)
            assert allclose(m.hyperparameters['sigma_n'], sigma_n[i], rtol=PARAM_RTOL)
    finally:
        clean(config)
        from os.path import splitext
        # .hdf5 is not yet supported for multi output model, so it is saved as .pkl instead.
        model_file = splitext(model_file)[0] + '.pkl'
        if path.exists(model_file):
            remove(model_file)


def test_2D():
    """Test a Rosenbrock 2D function with two random inputs."""

    config_file = 'study_2D/profit_2D.yaml'
    config = BaseConfig.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'GPy'
        assert sur.trained
        assert sur.model.kern.name == 'rbf'
        assert sur.model.kern.input_dim == 2
        assert allclose(sur.hyperparameters['length_scale'], 0.47321765, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_f'], 0.49950325, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_n'], 3.36370612e-07, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)


def test_2D_independent():
    """Test a Fermi function which returns a vector over energy and is sampled over different temperatures."""

    config_file = 'study_independent/profit_independent.yaml'
    config = BaseConfig.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'GPy'
        assert sur.trained
        assert sur.model.kern.name == 'rbf'
        assert sur.model.kern.input_dim == 1
        assert allclose(sur.hyperparameters['length_scale'], 0.25102422, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_f'], 0.36038181, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_n'], 0.0042839, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)


def test_karhunenloeve():
    """Same test function as 'test_2D_independent' but with multi-output surrogate and Karhunen-Loeve encoder."""

    config_file = 'study_karhunenloeve/profit_karhunenloeve.yaml'
    config = BaseConfig.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'CoregionalizedGPy'
        assert sur.trained
        assert allclose(sur.hyperparameters['length_scale'], 0.51021744, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_f'], 0.36038181, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_n'], 0.1267644, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            from os.path import splitext
            model_file = splitext(model_file)[0] + '.pkl'
            remove(model_file)



"""Global default configuration values."""
from os import path, getcwd

# Base Config
base_dir = path.abspath(getcwd())
run_dir = base_dir
config_file = 'profit.yaml'
include = []
files = {'input': 'input.txt',
         'output': 'output.txt'}
ntrain = 10
variables = {}

# Run Config
run = {'runner': 'local',
       'interface': 'memmap',
       'pre': 'template',
       'post': 'json',
       'command': './simulation',
       'stdout': 'stdout',
       'stderr': None,
       'clean': True,
       'time': True,
       'debug': False,
       'log_path': 'log',
       'custom': False,
       'worker': None}


run_runner_local = {'class': 'local',
                    'parallel': 'all',
                    'sleep': 0,
                    'fork': True}

run_runner_slurm = {'class': 'slurm',
                    'parallel': None,
                    'sleep': 0,
                    'poll': 60,
                    'path': 'slurm.bash',
                    'custom': False,
                    'prefix': 'srun',
                    'OpenMP': False,
                    'cpus': 1,
                    'options': {'job-name': 'profit'}}

run_interface_memmap = {'class': 'memmap',
                        'path': 'interface.npy'}

run_interface_zeromq = {'class': 'zeromq',
                        'transport': 'tcp',
                        'port': 9000,
                        'address': None,
                        'connect': None,
                        'timeout': 2500,
                        'retries': 3,
                        'retry-sleep': 1}

run_pre_template = {'class': 'template',
                    'path': 'template',
                    'param_files': None}

run_post_json = {'class': 'json',
                 'path': 'stdout'}

run_post_numpytxt = {'class': 'numpytxt',
                     'path': 'stdout',
                     'names': 'all',
                     'options': {'deletechars': ""}}

run_post_hdf5 = {'class': 'hdf5',
                 'path': 'output.hdf5'}


# Fit Config
fit = {'surrogate': 'GPy',
       'save': './model.hdf5',
       'load': False,
       'fixed_sigma_n': False,
       'encoder': ['Exclude(Constant)', 'Log10(LogUniform)', 'Normalization(All)'],
       # for internal use:
       '_input_encoders': [],
       '_output_encoders': [],
       }

fit_gaussian_process = {'surrogate': 'GPy',
                        'kernel': 'RBF',
                        'hyperparameters': {'length_scale': None,  # Hyperparameters are inferred from training data
                                            'sigma_n': None,
                                            'sigma_f': None}}

# Active Learning Config
active_learning = {'algorithm': 'simple',
                   'nwarmup': 3,
                   'batch_size': 1,
                   'convergence_criterion': 1e-5,
                   'nsearch': 50,
                   'make_plot': False,
                   'save_intermediate': {'model_path': './model.hdf5',
                                         'input_path': './input.txt',
                                         './output_path': 'output.txt'},
                   'resume_from': None
                   }

al_algorithm_simple = {'class': 'simple',
                       'acquisition_function': 'simple_exploration',
                       'save': True}
al_algorithm_mcmc = {'class': 'mcmc',
                     'reference_data': './yref.txt',
                     'warmup_cycles': 1,
                     'target_acceptance_rate': 0.35,
                     'sigma_n': 0.05,
                     'initial_points': None,
                     'last_percent': 0.25,
                     'save': './mcmc_model.hdf5',
                     'delayed_acceptance': False}

al_acquisition_function_simple_exploration = {'class': 'simple_exploration',
                                              'use_marginal_variance': False}
al_acquisition_function_exploration_with_distance_penalty = {'class': 'exploration_with_distance_penalty',
                                                             'use_marginal_variance': False,
                                                             'weight': 10}
al_acquisition_function_weighted_exploration = {'class': 'weighted_exploration',
                                                'use_marginal_variance': False,
                                                'weight': 0.5}
al_acquisition_function_probability_of_improvement = {'class': 'probability_of_improvement'}
al_acquisition_function_expected_improvement = {'class': 'expected_improvement',
                                                'exploration_factor': 0.01,
                                                'find_min': False}
al_acquisition_function_expected_improvement_2 = {'class': 'expected_improvement_2',
                                                  'exploration_factor': 0.01,
                                                  'find_min': False}
al_acquisition_function_alternating_exploration = {'class': 'alternating_exploration',
                                                   'use_marginal_variance': False,
                                                   'exploration_factor': 0.01,
                                                   'find_min': False,
                                                   'alternating_freq': 1}


# UI Config
ui = {'plot': False}



"""
Testcases for variables:
- Input
- Output
- ActiveLearning
- VariableGroup
"""
import numpy as np
from profit.util.variable import Variable


def test_input():
    from profit.util.variable import InputVariable
    NAME = 'v'
    KIND = 'uniform'
    CONSTRAINTS = (2, 5)
    SIZE = (3, 1)
    DTYPE = np.float64
    VARIABLE_DICT = {'name': NAME, 'kind': KIND, 'constraints': CONSTRAINTS,
                     'size': SIZE, 'value': None, 'dtype': DTYPE}

    v1 = Variable.create(NAME, KIND, SIZE, constraints=CONSTRAINTS)
    constraints_str = ', '.join(str(c) for c in CONSTRAINTS)
    v2 = Variable.create_from_str(NAME, SIZE, "{kind}({constraints})".format(kind=KIND, constraints=constraints_str))
    v3 = InputVariable(NAME, KIND, SIZE, CONSTRAINTS)
    variables = (v1, v2, v3)
    assert all(v.name == NAME for v in variables)
    assert all(v.size == SIZE for v in variables)
    assert all(v.dtype == DTYPE for v in variables)
    for v in variables:
        v.generate_values()
    VARIABLE_DICT['value'] = variables[0].value
    assert all(CONSTRAINTS[0] <= v.value.min() and v.value.max() <= CONSTRAINTS[1] for v in variables)
    assert np.all(v.named_value == np.array(v.value, dtype=[(v.name, v.dtype)]) for v in variables)
    assert np.all(v.as_dict() == VARIABLE_DICT for v in variables)
    assert np.all(v[key] == value for key, value in VARIABLE_DICT.items() for v in variables)

    CONST = 1.4
    v4 = Variable.create_from_str(NAME, SIZE, f"{CONST}")
    assert v4.kind.lower() == 'constant'
    v4.generate_values()
    assert np.all(v4.value == np.ones(SIZE) * CONST)


def test_output():
    from profit.util.variable import OutputVariable
    NAME = 'f'
    KIND = 'Output'
    DEPENDENT = 't'
    SIZE = (3, 1)
    DTYPE = np.float64
    VARIABLE_DICT = {'name': NAME, 'kind': KIND, 'dependent': DEPENDENT,
                     'size': SIZE, 'value': None, 'dtype': DTYPE}

    DEPENDENT_SIZE = (200, 1)
    DEPENDENT_VARIABLE = Variable.create_from_str('t', DEPENDENT_SIZE, "Independent(0, 100, 200)")
    v1 = Variable.create(NAME, KIND, SIZE, dependent=DEPENDENT_VARIABLE)
    v2 = Variable.create_from_str(NAME, SIZE, f"{KIND}({DEPENDENT})")
    v3 = OutputVariable(NAME, KIND, SIZE, DEPENDENT)
    variables = (v1, v2, v3)
    for v in variables:
        v.resolve_dependent(DEPENDENT_VARIABLE)
    assert all(v.name == NAME for v in variables)
    assert all(v.size == (SIZE[0], DEPENDENT_SIZE[0]) for v in variables)
    assert all(v.dtype == DTYPE for v in variables)
    VARIABLE_DICT['value'] = variables[0].value
    assert all(v.dependent[0].as_dict() == DEPENDENT_VARIABLE.as_dict() for v in variables)
    assert np.all(v.named_value == np.array(v.value, dtype=[(v.name, v.dtype)]) for v in variables)
    assert np.all(v.as_dict() == VARIABLE_DICT for v in variables)
    assert np.all(v[key] == value for key, value in VARIABLE_DICT.items() for v in variables)


def test_active_learning():
    from profit.util.variable import ActiveLearningVariable
    NAME = 'v'
    KIND = 'ActiveLearning'
    DISTR = 'uniform'
    CONSTRAINTS = (0, 2)
    SIZE = (10, 1)
    DTYPE = np.float64
    VARIABLE_DICT = {'name': NAME, 'kind': KIND, 'distr': DISTR, 'constraints': CONSTRAINTS,
                     'size': SIZE, 'value': None, 'dtype': DTYPE}

    v1 = Variable.create(NAME, KIND, SIZE, distr=DISTR, constraints=CONSTRAINTS)
    constraints_str = ', '.join(str(c) for c in CONSTRAINTS)
    v2 = Variable.create_from_str(NAME, SIZE, "{kind}({constraints})".format(kind=KIND, constraints=constraints_str))
    v3 = ActiveLearningVariable(NAME, KIND, SIZE, DISTR, CONSTRAINTS)
    variables = (v1, v2, v3)

    assert all(v.name == NAME for v in variables)
    assert all(v.size == SIZE for v in variables)
    assert all(v.dtype == DTYPE for v in variables)

    VARIABLE_DICT['value'] = variables[0].value
    assert all(v.distr == DISTR for v in variables)
    assert np.all(v.named_value == np.array(v.value, dtype=[(v.name, v.dtype)]) for v in variables)
    assert np.all(v.as_dict() == VARIABLE_DICT for v in variables)
    assert np.all(v[key] == value for key, value in VARIABLE_DICT.items() for v in variables)

    # Distribute prediction points for AL according to distribution
    Xpred = v1.create_Xpred(SIZE)
    assert np.all(Xpred == np.linspace(*CONSTRAINTS, SIZE[0]).reshape(SIZE))

    DISTR = 'log'
    CONSTRAINTS = (1e-2, 1e3)
    v4 = ActiveLearningVariable(NAME, KIND, SIZE, DISTR, CONSTRAINTS)
    Xpred = v4.create_Xpred(SIZE)
    assert np.all(Xpred == CONSTRAINTS[0] * np.exp((np.log(CONSTRAINTS[1]) - np.log(CONSTRAINTS[0]))
                                                           * np.linspace(0, 1, SIZE[0])).reshape(SIZE))


def test_variablegroup():
    from profit.util.variable import VariableGroup
    SIZE = (3, 1)
    var = {'v': 'Uniform(0, 1)',
           'u': 'ActiveLearning(3, 4)',
           't': 'Independent(1, 5, 100)',
           'f': 'Output(t)'}

    variables = [Variable.create_from_str(name, SIZE, v_str) for name, v_str in var.items()]
    group = VariableGroup(samples=SIZE[0])
    group.add(variables)

    assert all(group[v.name] is v for v in variables)
    # assert group.all  # Not working yet
    variables[1].value = np.random.random(SIZE)
    variables[-1].value = np.random.random(variables[-1].size)

    assert all(np.all(group.input[:, i] == v.value.flat) for i, v in enumerate(variables[:2]))
    assert all(np.all(group.named_input[v.name] == v.value) for v in variables[:2])
    assert np.all(group.output == variables[-1].value)
    assert np.all(group.named_output['f'] == variables[-1].value)
    assert group.formatted_output.dtype.names == tuple(f'f({t})' for t in variables[2].value.flat)
    assert all(group.as_dict[v.name] == v.as_dict() for v in variables)
    group.delete_variable(0)
    assert all(np.all(group.input[:, i] == v.value.flat) for i, v in enumerate(variables[1:2]))
    group.delete_sample(0)
    assert group.input.shape[0] == group.output.shape[0] == \
           variables[1].value.shape[0] == variables[-1].value.shape[0] == SIZE[0] - 1




"""
* simple exploration
    - variance
    - variance + penalty for near distance to previous point
    - weighted exploration/exploitation: weight * mu + (1 - weight) * sigma
* bayesian optimization
    - probability of improvement
    - expected improvement
* mixed exploration and bayesian optimization
"""

from profit.util.base_class import CustomABC
import numpy as np

from profit.defaults import \
    al_acquisition_function_simple_exploration as se_defaults, \
    al_acquisition_function_exploration_with_distance_penalty as edp_defaults, \
    al_acquisition_function_weighted_exploration as we_defaults, \
    al_acquisition_function_probability_of_improvement as poi_defaults, \
    al_acquisition_function_expected_improvement as ei_defaults, \
    al_acquisition_function_expected_improvement_2 as ei2_defaults, \
    al_acquisition_function_alternating_exploration as ae_defaults


class AcquisitionFunction(CustomABC):
    """Base class for acquisition functions.

    Parameters:
        Xpred (np.ndarray): Matrix of possible training points.
        surrogate (profit.sur.Surrogate): Surrogate.
        variables (profit.util.variable.VariableGroup): Variables.
        parameters (dict): Miscellaneous parameters for the specified function. E.g. 'exploration_factor'.
    """
    labels = {}
    al_parameters = {}

    EPSILON = 1e-12

    def __init__(self, Xpred, surrogate, variables, **parameters):
        self.parameters = parameters

        self.Xpred = Xpred
        self.surrogate = surrogate
        self.variables = variables

    def set_al_parameters(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.al_parameters:
                self.al_parameters[key] = value
            else:
                print(f"Skipped setting AL parameter {key}.")

    def calculate_loss(self, *args):
        """Calculate the loss of the acquisition function."""
        return np.full(self.Xpred.shape[0], np.nan)

    def find_next_candidates(self, batch_size):
        """Find the next training input points which minimize the loss/maximize improvement."""
        return self._find_next_candidates(batch_size)

    def _find_next_candidates(self, batch_size, *loss_args):
        candidates = np.empty((batch_size, self.Xpred.shape[-1]))
        mask = np.ones(self.Xpred.shape[0])
        for n in range(batch_size):
            loss = self.calculate_loss(*loss_args)
            idx = np.argmax(loss * mask)
            mask[idx] = 0  # Exclude already visited points.
            candidates[n] = self.Xpred[idx.flatten()]
            mu_candidate, _ = self.surrogate.predict(candidates[n].reshape(1, -1))
            self.surrogate.add_training_data(candidates[n].reshape(1, -1), mu_candidate)
            self.surrogate.optimize()
        return candidates

    def normalize(self, value, min=None):
        minval = value.min(axis=0)
        maxval = value.max(axis=0)
        normalized_value = (value - minval) / np.maximum((maxval - minval), self.EPSILON)
        if min is not None:
            return np.maximum(normalized_value, min)
        return normalized_value


@AcquisitionFunction.register("simple_exploration")
class SimpleExploration(AcquisitionFunction):
    """Minimizes the local variance, which means the next points are generated at points of high variance."""

    def __init__(self, Xpred, surrogate, variables, use_marginal_variance=se_defaults['use_marginal_variance'], **parameters):
        super().__init__(Xpred, surrogate, variables, use_marginal_variance=use_marginal_variance, **parameters)

    def calculate_loss(self):
        if self.parameters['use_marginal_variance']:
            if hasattr(self.surrogate, 'get_marginal_variance'):
                variance = self.surrogate.get_marginal_variance(self.Xpred)
            else:
                print("Surrogate has no method 'get_marginal_variance'. Using predictive variance instead.")
                _, variance = self.surrogate.predict(self.Xpred)
        else:
            _, variance = self.surrogate.predict(self.Xpred)
        loss = np.sum(variance, axis=1)
        return loss


@AcquisitionFunction.register("exploration_with_distance_penalty")
class ExplorationWithDistancePenalty(SimpleExploration):
    r"""Enhanced variance minimization by adding an exponential penalty for neighboring candidates.

    Variables:
        weight (float): Exponential penalty factor: $penalty = 1 - exp(c1 * |X_{pred} - X_{last}|)$.
    """

    def __init__(self, Xpred, surrogate, variables, use_marginal_variance=edp_defaults['use_marginal_variance'],
                 weight=edp_defaults['weight']):
        super().__init__(Xpred, surrogate, variables, use_marginal_variance=use_marginal_variance, weight=weight)

    def calculate_loss(self):
        c1 = self.parameters['weight']
        loss = super().calculate_loss()
        loss_scale = loss.max(axis=0)
        last_point = self.variables.input[np.sum(~np.isnan(self.variables.input), axis=0).min() - 1]
        loss += loss_scale * (1.0 - np.exp(-c1 * np.linalg.norm(self.Xpred - last_point, axis=1)))
        loss /= 2
        return loss


@AcquisitionFunction.register("weighted_exploration")
class WeightedExploration(AcquisitionFunction):
    """Combination of exploration and optimization.

    Variables:
        weight (float): Factor to favor maximization of the target function over exploration.
    """

    def __init__(self, Xpred, surrogate, variables, weight=we_defaults['weight'],
                 use_marginal_variance=we_defaults['use_marginal_variance']):
        super().__init__(Xpred, surrogate, variables, weight=weight, use_marginal_variance=use_marginal_variance)

    def calculate_loss(self, mu):
        weight = self.parameters['weight']

        _, variance = self.surrogate.predict(self.Xpred)
        if self.parameters['use_marginal_variance']:
            if hasattr(self.surrogate, 'get_marginal_variance'):
                variance = self.surrogate.get_marginal_variance(self.Xpred)
            else:
                print("Surrogate has no method 'get_marginal_variance'. Using predictive variance instead.")
        variance = self.normalize(variance)
        loss = weight * mu + (1 - weight) * variance
        loss = np.sum(loss, axis=1)
        return loss

    def find_next_candidates(self, batch_size):
        mu, _ = self.surrogate.predict(self.Xpred)
        mu = self.normalize(mu)
        return self._find_next_candidates(batch_size, mu)


@AcquisitionFunction.register("probability_of_improvement")
class ProbabilityOfImprovement(AcquisitionFunction):
    """Maximizes the probability of improvement.
    See https://math.stackexchange.com/questions/4230985/probability-of-improvement-pi-acquisition-function-for-bayesian-optimization
    """

    def calculate_loss(self, mu):
        _, variance = self.surrogate.predict(self.Xpred)
        improvement = np.maximum(mu + np.sqrt(variance) * np.random.standard_normal(mu.shape) -
                                 self.variables.output.max(axis=0), 0)
        return np.sum(improvement, axis=1)

    def find_next_candidates(self, batch_size):
        mu, _ = self.surrogate.predict(self.Xpred)
        return self._find_next_candidates(batch_size, mu)


@AcquisitionFunction.register("expected_improvement")
class ExpectedImprovement(AcquisitionFunction):
    """Maximising the expected improvement.
    See https://krasserm.github.io/2018/03/21/bayesian-optimization/

    To be able to execute this funciton with batches of data, some simplifications are made:
    The optimization part (prediction mean) is only calculated once for the first point. Thereafter, it is assumed
    that the data coincides with the prediction. For the next points in the batch, only the variance part is calculated
    as this does not need an evaluation of the function.
    """

    SIGMA_EPSILON = 1e-10

    def __init__(self, Xpred, surrogate, variables, exploration_factor=ei_defaults['exploration_factor'],
                 find_min=ei_defaults['find_min']):
        super().__init__(Xpred, surrogate, variables, exploration_factor=exploration_factor, find_min=find_min)

    def calculate_loss(self, improvement):
        from scipy.stats import norm

        sigma = self.sigma_part()
        z = improvement / sigma
        expected_improvement = improvement * norm.cdf(z) + sigma * norm.pdf(z)
        return np.sum(expected_improvement, axis=1)

    def mu_part(self):
        xi = self.parameters['exploration_factor']
        mu, _ = self.surrogate.predict(self.Xpred)
        if self.parameters['find_min']:
            mu *= -1
        improvement = np.maximum(mu - np.nanmax(self.variables.output, axis=0), 0)
        improvement = self.normalize(improvement) - xi
        return improvement

    def sigma_part(self):
        _, variance = self.surrogate.predict(self.Xpred)
        sigma = np.sqrt(variance)
        sigma = self.normalize(sigma, min=self.SIGMA_EPSILON)
        return sigma

    def find_next_candidates(self, batch_size):
        improvement = self.mu_part()
        return self._find_next_candidates(batch_size, improvement)


@AcquisitionFunction.register("expected_improvement_2")
class ExpectedImprovement2(AcquisitionFunction):
    """Simplified batch expected improvement where the first point is calculated using normal expected improvement,
    while the others are found using the minimization of local variance acquisition function.
    """

    def __init__(self, Xpred, surrogate, variables, exploration_factor=ei2_defaults['exploration_factor'],
                 find_min=ei2_defaults['find_min']):
        super().__init__(Xpred, surrogate, variables, exploration_factor=exploration_factor, find_min=find_min)

    def calculate_loss(self):
        from scipy.stats import norm
        xi = self.parameters['exploration_factor']
        mu, variance = self.surrogate.predict(self.Xpred)
        if self.parameters['find_min']:
            mu *= -1
        sigma = np.maximum(np.sqrt(variance), 1e-10)
        improvement = mu - np.nanmax(self.variables.output, axis=0) - xi
        z = improvement / sigma
        expected_improvement = improvement * norm.cdf(z) + sigma * norm.pdf(z)
        expected_improvement[sigma == 0] = 0
        return np.sum(expected_improvement, axis=1)

    def find_next_candidates(self, batch_size):
        candidates = np.empty((batch_size, self.Xpred.shape[-1]))
        loss = self.calculate_loss()
        idx = np.argmax(loss)
        candidates[0] = self.Xpred[idx.flatten()]
        mu_candidate = self.surrogate.predict(candidates[0].reshape(1, -1))
        self.surrogate.add_training_data(candidates[0].reshape(1, -1), mu_candidate)

        simple_exploration = SimpleExploration(self.Xpred, self.surrogate, self.variables)
        for n in range(1, batch_size):
            loss = simple_exploration.calculate_loss()
            idx = np.argmax(loss)
            candidates[n] = self.Xpred[idx.flatten()]
            mu_candidate = self.surrogate.predict(candidates[n].reshape(1, -1))
            simple_exploration.surrogate.add_training_data(candidates[n].reshape(1, -1), mu_candidate)
        return candidates


@AcquisitionFunction.register("alternating_exploration")
class AlternatingAF(AcquisitionFunction):

    al_parameters = {'krun': 0}

    def __init__(self, Xpred, surrogate, variables, use_marginal_variance=ae_defaults['use_marginal_variance'],
                 exploration_factor=ae_defaults['exploration_factor'],
                 find_min=ae_defaults['find_min'], alternating_freq=ae_defaults['alternating_freq']):
        super().__init__(Xpred, surrogate, variables, alternating_freq=alternating_freq)
        self.exploration = SimpleExploration(Xpred, surrogate, variables, use_marginal_variance=use_marginal_variance)
        self.expected_improvement = ExpectedImprovement(Xpred, surrogate, variables,
                                                        exploration_factor=exploration_factor, find_min=find_min)
        self.current_af = self.exploration

    def find_next_candidates(self, batch_size):
        if not self.al_parameters['krun'] % self.parameters['alternating_freq']:
            if self.current_af == self.exploration:
                self.current_af = self.expected_improvement
            else:
                self.current_af = self.exploration
        candidates = self.current_af.find_next_candidates(batch_size)
        return candidates



from os import path
import yaml
from collections import OrderedDict
from profit import defaults
from profit.util.base_class import CustomABC
import warnings

VALID_FORMATS = ('.yaml', '.py')

"""
yaml has to be configured to represent OrderedDict 
see https://stackoverflow.com/questions/16782112/can-pyyaml-dump-dict-items-in-non-alphabetical-order
and https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
"""


def represent_ordereddict(dumper, data):
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)


def dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
yaml.add_representer(OrderedDict, represent_ordereddict)
yaml.add_constructor(_mapping_tag, dict_constructor)

""" now yaml is configured to handle OrderedDict input and output """


def load_config_from_py(filename):
    """ Load the configuration parameters from a python file into dict. """
    from importlib.util import spec_from_file_location, module_from_spec
    spec = spec_from_file_location('f', filename)
    f = module_from_spec(spec)
    spec.loader.exec_module(f)
    return {name: value for name, value in f.__dict__.items() if not name.startswith('_')}


class AbstractConfig(CustomABC):
    """General class with methods which are useful for all Config classes."""
    labels = {}
    defaults = None

    def __init__(self, **entries):
        if self.defaults:
            self.set_defaults(getattr(defaults, self.defaults))
        self.update(**entries)

    def update(self, **entries):
        """Updates the attributes with user inputs. A warning is issued if the attribute set by the user is unknown.

        Parameters:
            entries (dict): User input of the config parameters.
        """
        for name, value in entries.items():
            if hasattr(self, name) or name in map(str.lower, self.labels):
                attr = getattr(self, name, None)
                if isinstance(attr, dict):
                    attr.update(value)
                    setattr(self, name, attr)
                else:
                    setattr(self, name, value)
            else:
                message = f"Config parameter '{name}' for {self.__class__.__name__} configuration may be unused."
                warnings.warn(message)
                setattr(self, name, value)

    def process_entries(self, base_config):
        """After the attributes are set, they are formatted and edited to standardize the user inputs.

        Parameters:
            base_config (BaseConfig): In sub configs, the data from the base config is needed.
        """
        pass

    def set_defaults(self, default_dict):
        """Default values are set from a default dictionary, which is usually located
        in the global profit.defaults file.
        """
        for name, value in default_dict.items():
            if name in self.labels and isinstance(value, str):
                value = {'class': value}
            setattr(self, name, value)

    def create_subconfig(self, sub_config_label, **entries):
        """Instances of sub configs are created from a string or a dictionary.

        Parameters:
            sub_config_label (str): Dict key of registered sub config.
            entries (dict): User input parameters.
        """
        if 'class' in entries:
            # Load specific sub config or default config, if missing.
            try:
                sub = self.labels[sub_config_label][entries['class']]()
            except KeyError:
                sub = self.labels[sub_config_label]['default'](**entries)
        else:
            # Load general sub config.
            sub = self.labels[sub_config_label]()

        # Split entries into entries for this config and further sub configs.
        base_entries = {k: v for k, v in entries.items() if k.lower() not in sub.labels}
        sub_entries = {k: {'class': v} if isinstance(v, str) else v
                       for k, v in entries.items() if k.lower() in sub.labels}

        # Update defaults with user entries
        sub.update(**base_entries)

        # Create second level sub configs.
        for subsub_label in sub.labels:
            subsub_entries = sub[subsub_label]
            subsub_entries.update(sub_entries.get(subsub_label, {}))
            sub.create_subconfig(subsub_label, **subsub_entries)
        setattr(self, sub_config_label, sub)

    def __getitem__(self, item):
        """Implements the dictionary like get method with brackets.

        Parameters:
            item (str): Label of the attribute to return.

        Returns:
            Attribute or if the attribute is a sub config, a dictionary of the sub config items.
        """
        attr = getattr(self, item)
        if item in self.labels:
            if type(attr) is list:
                return {"list": attr}
            return {key: attr[key] for key, _ in attr.items()}
        return attr

    def items(self):
        """Implements the dictionary like self.items() method.

        Returns:
            list: List of (key, value) tuples of the class attributes.
        """
        return [(key, self[key]) for key in vars(self)]

    def get(self, item, default=None):
        """Implements the dictionary like get method with a default value.

        Parameters:
            item (str): Label of the attribute to return.
            default: Default value, if the attribute is not found.

        Returns:
            Attribute or the default value.
        """
        try:
            return self[item]
        except AttributeError:
            return default


class BaseConfig(AbstractConfig):
    """
    This class and its modular subclasses provide all possible configuration parameters.

    Parts of the Config:
        - base_dir
        - run_dir
        - config_file
        - include
        - ntrain
        - variables
        - files
            - input
            - output
        - run
            - runner
            - interface
            - pre
            - post
        - fit
            - surrogate
            - save / load
            - fixed_sigma_n
        - active_learning
        - ui

    Base configuration for fundamental parameters.

    Parameters:
        base_dir (str): Base directory.
        run_dir (str): Run directory.
        config_path (str): Path to configuration file.
        include (list): Paths to custom files which are loaded in the beginning.
        files (dict): Paths for input and output files.
        ntrain (int): Number of training samples.
        variables (dict): All variables.
        input (dict): Input variables.
        output (dict): Output variables.
        independent (dict): Independent variables, if the result of the simulation is a vector.
    """
    labels = {}

    def __init__(self, base_dir=defaults.base_dir, **entries):
        # Set defaults
        self.base_dir = path.abspath(base_dir)
        self.run_dir = self.base_dir
        self.config_path = path.join(self.base_dir, defaults.config_file)
        self.include = defaults.include
        self.ntrain = defaults.ntrain
        self.variables = defaults.variables.copy()
        self.input = {}
        self.output = {}
        self.independent = {}
        self.files = defaults.files.copy()

        # Split user entries in entries for base_config and for sub_configs
        base_entries = {k: v for k, v in entries.items() if k.lower() not in self.labels}
        sub_entries = {k: {'class': v} if isinstance(v, str) else v
                       for k, v in entries.items() if k.lower() in self.labels}

        self.update(**base_entries)  # Update the attributes with given entries.
        self.load_includes()  # Load external files.

        for sub_config_label in self.labels:
            single_sub_entries = sub_entries.get(sub_config_label, {})
            self.create_subconfig(sub_config_label, **single_sub_entries)

        self.process_entries()  # Postprocess the attributes to standardize different user entries.

    def process_entries(self):
        """Sets absolute paths, creates variables and delegates to the sub configs."""
        from profit.util.variable import Variable, VariableGroup

        # Set absolute paths
        self.files['input'] = path.join(self.base_dir, self.files.get('input', defaults.files['input']))
        self.files['output'] = path.join(self.base_dir, self.files.get('output', defaults.files['output']))

        # Variable configuration as dict
        self.variable_group = VariableGroup(self.ntrain)
        vars = []
        for k, v in self.variables.items():
            if isinstance(v, (int, float)):
                v = f'Constant({v})'
            if isinstance(v, str):
                vars.append(Variable.create_from_str(k, (self.ntrain, 1), v))
            else:
                vars.append(Variable.create(name=k, size=(self.ntrain,1), **v))
        self.variable_group.add(vars)

        self.variables = self.variable_group.as_dict
        self.input = {k: v for k, v in self.variables.items()
                      if not any(k in v['kind'].lower() for k in ('output', 'independent'))}
        self.output = {k: v for k, v in self.variables.items()
                       if 'output' in v['kind'].lower()}
        self.independent = {k: v for k, v in self.variables.items()
                            if 'independent' in v['kind'].lower() and v['size'] != (1, 1)}

        for sub_config_label in self.labels:
            getattr(self, sub_config_label).process_entries(self)

    @classmethod
    def from_file(cls, filename=defaults.config_file):
        """Creates a configuration class from a .yaml or .py file."""

        if filename.endswith('.yaml'):
            with open(filename) as f:
                entries = yaml.safe_load(f)
        elif filename.endswith('.py'):
            entries = load_config_from_py(filename)
        else:
            raise TypeError(f"Not supported file extension .{filename.split('.')[-1]} for config file.\n"
                            f"Valid file formats: {VALID_FORMATS}")
        self = cls(base_dir=path.split(filename)[0], **entries)
        self.config_path = path.join(self.base_dir, filename)
        return self

    def load_includes(self):
        from profit.util import load_includes

        if isinstance(self.include, str):
            self.include = [self.include]

        self.include = [path.abspath(path.join(self.base_dir, p)) for p in self.include]
        load_includes(self.include)


@BaseConfig.register("run")
class RunConfig(AbstractConfig):
    """Run configuration with the following sub classes:
        - runner
            - local
            - slurm
        - interface
            - memmap
            - zeromq
        - pre
            - template
        - post
            - json
            - numpytxt
            - hdf5

    A default sub class which just updates the entries from a user input is also implemented and used if the
    class from the user input is not found.

    Custom config classes can also be registered, e.g. as a custom runner:

    .. code-block:: python

        @RunnerConfig.register("custom")
        class CustomRunner(LocalRunnerConfig):
            def process_entries(self, base_config):
                # do something else than the usual LocalRunnerConfig
                pass

    Default values from the global profit.defaults.py file are loaded.
    """
    labels = {}
    defaults = "run"

    def process_entries(self, base_config):
        """Set paths and process entries of sub configs."""

        if not path.isabs(self.log_path):
            self.log_path = path.abspath(path.join(base_config.base_dir, self.log_path))

        for key in self.labels:
            getattr(self, key.lower()).process_entries(base_config)


@RunConfig.register("runner")
class RunnerConfig(AbstractConfig):
    """Base Runner config."""
    labels = {}
    defaults = None


@RunnerConfig.register("local")
class LocalRunnerConfig(RunnerConfig):
    """
    Example:
        .. code-block:: yaml

            class: local
            parallel: all   # maximum number of simultaneous runs (for spawn array)
            sleep: 0        # number of seconds to sleep while polling
            fork: true      # whether to spawn the worker via forking instead of a subprocess (via a shell)
    """
    labels = {}
    defaults = "run_runner_local"

    def process_entries(self, base_config):
        """Converts `parallel: all` to number of available cpus"""
        from os import sched_getaffinity
        if self.parallel == 'all':
            self.parallel = len(sched_getaffinity(0))


@RunnerConfig.register("slurm")
class SlurmRunnerConfig(RunnerConfig):
    """
    Example:
        .. code-block:: yaml

               class: slurm
               parallel: null      # maximum number of simultaneous runs (for spawn array)
               sleep: 0            # number of seconds to sleep while (internally) polling
               poll: 60            # number of seconds between external polls (to catch failed runs), use with care!
               path: slurm.bash    # the path to the generated batch script (relative to the base directory)
               custom: false       # whether a custom batch script is already provided at 'path'
               prefix: srun        # prefix for the command
               OpenMP: false       # whether to set OMP_NUM_THREADS and OMP_PLACES
               cpus: 1             # number of cpus (including hardware threads) to use (may specify 'all')
               options:            # (long) options to be passed to slurm: e.g. time, mem-per-cpu, account, constraint
                   job-name: profit
    """
    labels = {}
    defaults = "run_runner_slurm"

    def process_entries(self, base_config):
        """Converts paths to absolute and check type of 'cpus'"""
        # Convert path to absolute path
        if not path.isabs(self.path):
            self.path = path.abspath(path.join(base_config.base_dir, self.path))
        # Check type of 'cpus'
        if (type(self.cpus) is not int or self.cpus < 1) and self.cpus != 'all':
            raise ValueError(f'config option "cpus" may only be a positive integer or "all" and not {self.cpus}')


@RunConfig.register("interface")
class InterfaceConfig(AbstractConfig):
    """Base runner interface config."""
    labels = {}
    defaults = None


@InterfaceConfig.register("memmap")
class MemmapInterfaceConfig(InterfaceConfig):
    """
    Example:
        .. code-block:: yaml

            class: memmap
            path: interface.npy     # path to memory mapped interface file, relative to base directory
    """
    labels = {}
    defaults = "run_interface_memmap"

    def process_entries(self, base_config):
        """Converts 'path' to absolute."""
        if not path.isabs(self.path):
            self.path = path.abspath(path.join(base_config.base_dir, self.path))


@InterfaceConfig.register("zeromq")
class ZeroMQInterfaceConfig(InterfaceConfig):
    """
    Example:
        .. code-block:: yaml

            class: zeromq
            transport: tcp      # transport system used by zeromq
            port: 9000          # port for the interface
            address: null       # override bind address used by zeromq
            connect: null       # override connect address used by zeromq
            timeout: 2500       # zeromq polling timeout, in ms
            retries: 3          # number of zeromq connection retries
            retry-sleep: 1      # sleep between retries, in s
    """
    labels = {}
    defaults = "run_interface_zeromq"


@RunConfig.register("pre")
class PreConfig(AbstractConfig):
    """Base config for preprocessors."""
    labels = {}
    defaults = None


@PreConfig.register("template")
class TemplatePreConfig(PreConfig):
    """
    Example:
        .. code-block:: yaml

            class: template
            path: template      # directory to copy from, relative to base directory
            param_files: null   # files in template which contain placeholders for variables, null means all files
                                # can be a filename or a list of filenames
    """
    labels = {}
    defaults = "run_pre_template"

    def process_entries(self, base_config):
        """Convert 'path' to absolute and set 'param_files'."""
        if not path.isabs(self.path):
            self.path = path.abspath(path.join(base_config.base_dir, self.path))

        if isinstance(self.param_files, str):
            self.param_files = [self.param_files]


@RunConfig.register("post")
class PostConfig(AbstractConfig):
    """Base class for postprocessor configs."""
    labels = {}
    defaults = None


@PostConfig.register("json")
class JsonPostConfig(PostConfig):
    """
    Example:
        .. code-block:: yaml

            class: json
            path: stdout    # file to read from, relative to the run directory
    """
    labels = {}
    defaults = "run_post_json"


@PostConfig.register("numpytxt")
class NumpytxtPostConfig(PostConfig):
    """
    Example:
        .. code-block:: yaml

            class: numpytxt
            path: stdout    # file to read from, relative to the run directory
            names: "f g"    # list or string of output variables in order, default read from config/variables
            options:        # options which are passed on to numpy.genfromtxt() (fname & dtype are used internally)
                deletechars: ""
    """
    labels = {}
    defaults = "run_post_numpytxt"

    def process_entries(self, base_config):
        """Sets the included names of variables. The Keyword 'all' includes all variables."""
        if isinstance(self.names, str):
            self.names = list(base_config.output.keys()) if self.names == 'all' else self.names.split()


@PostConfig.register("hdf5")
class HDF5PostConfig(PostConfig):
    """
    Example:
        .. code-block:: yaml

            class: hdf5
            path: output.hdf5   # file to read from, relative to the run directory
    """
    labels = {}
    defaults = "run_post_hdf5"


@BaseConfig.register("fit")
class FitConfig(AbstractConfig):
    """Configuration for the surrogate and encoder. Currently, the only sub config is for the GaussianProcess classes."""
    labels = {}
    defaults = "fit"

    def __init__(self, **entries):
        from profit.sur import Surrogate
        from profit.sur.gp.gaussian_process import GaussianProcess
        self.set_defaults(defaults.fit)

        if issubclass(Surrogate.labels[self.surrogate], GaussianProcess):
            self.set_defaults(defaults.fit_gaussian_process)

        self.update(**entries)

    def process_entries(self, base_config):
        """Set 'load' and 'save' as well as the encoder."""
        for mode_str in ('save', 'load'):
            filepath = getattr(self, mode_str)
            if filepath:
                if self.surrogate not in filepath:
                    filepath = filepath.rsplit('.', 1)
                    filepath = ''.join(filepath[:-1]) + f'_{self.surrogate}.' + filepath[-1]
                setattr(self, mode_str, path.abspath(path.join(base_config.base_dir, filepath)))

        if self.load:
            self.save = False

        # Encoders
        from re import match
        import numpy as np

        # array: which columns belong to which variables
        input_columns = np.array(sum(([var.name] * var.size[1]
                                      for var in base_config.variable_group.input_list), []))
        output_columns = np.array(sum(([var.name] * var.size[1]
                                       for var in base_config.variable_group.output_list), []))

        for config in self.encoder:
            # handle shorthand notation, e.g. Name(a,b) -> {class: Name, variables: [a, b]}
            if isinstance(config, str):
                try:
                    name, var_spec = match(r"(\w+)\((.*)\)", config).groups()
                except AttributeError as ex:
                    raise ValueError(f"unable to parse encoder shortcut <{config}>") from ex
                var_spec = [v.strip().lower() for v in var_spec.split(",")]  # variable specification
            elif isinstance(config, dict):
                name = config["class"]
                var_spec = [v.strip().lower() for v in config["variables"]]
            else:
                raise ValueError(f"unable to parse encoder <{config}>")

            # ToDo: check if var_spec is valid -> warn otherwise

            # select input columns based on variables or kinds
            if any(s in var_spec for s in ["all", "in", "input", "inputs"]):
                input_vars = base_config.variable_group.input_list
                input_select = np.arange(input_columns.size)
            else:
                input_vars = [var for var in base_config.variable_group.input_list
                              if var.name.lower() in var_spec or var.kind.lower() in var_spec]
                if input_vars:
                    input_select = np.hstack([np.arange(input_columns.size)[input_columns == var.name]
                                              for var in input_vars])
                else:
                    input_select = None

            # select output columns based on variable names or kinds
            if any(s in var_spec for s in ["all", "out", "output", "outputs"]):
                output_vars = base_config.variable_group.output_list
                output_select = np.arange(output_columns.size)
            else:
                output_vars = [var for var in base_config.variable_group.output_list
                               if var.name.lower() in var_spec or var.kind.lower() in var_spec]
                if output_vars:
                    output_select = np.hstack([np.arange(output_columns.size)[output_columns == var.name]
                                               for var in output_vars])
                else:
                    output_select = None

            # handle special cases
            if name == "Exclude":
                # remove excluded columns from column lists
                input_columns = np.array([c for c in input_columns if c not in (v.name for v in input_vars)])
                output_columns = np.array([c for c in output_columns if c not in (v.name for v in output_vars)])
            elif name in ["PCA", "KarhunenLoeve"]:
                # ToDo: can't handle dimensionality reduction yet
                if config is not self.encoder[-1]:
                    raise NotImplementedError("reduced dimensions cannot be encoded further")

            # add processed config to _input_encoders & _output_encoders
            for encoders, select in [(self._input_encoders, input_select), (self._output_encoders, output_select)]:
                if select is not None:
                    encoders.append({
                        "class": name,
                        "columns": select,
                        "parameters": {k: float(v) for k, v in config.get("parameters", {})}
                        if not isinstance(config, str) else {},
                    })


@BaseConfig.register("active_learning")
class ALConfig(AbstractConfig):
    """Active learning configuration."""
    labels = {}
    defaults = "active_learning"

    def process_entries(self, base_config):
        for key in self.labels:
            getattr(self, key.lower()).process_entries(base_config)


@ALConfig.register("algorithm")
class AlgorithmALConfig(AbstractConfig):
    labels = {}
    defaults = None


@AlgorithmALConfig.register("simple")
class SimpleALConfig(AlgorithmALConfig):
    labels = {}
    defaults = "al_algorithm_simple"

    def process_entries(self, base_config):
        if self.save:
            self.save = base_config['fit']['save']
        for sub_config_label in self.labels:
            getattr(self, sub_config_label).process_entries(base_config)


@AlgorithmALConfig.register("mcmc")
class McmcConfig(AlgorithmALConfig):
    labels = {}
    defaults = "al_algorithm_mcmc"

    def process_entries(self, base_config):
        self.save = path.abspath(path.join(base_config.base_dir, self.save))
        self.reference_data = path.abspath(path.join(base_config.base_dir, self.reference_data))


@SimpleALConfig.register("acquisition_function")
class AcquisitionFunctionConfig(AbstractConfig):
    """Acquisition function configuration."""
    labels = {}
    defaults = None

    def process_entries(self, base_config):
        for k, v in self.items():
            if isinstance(v, str):
                try:
                    setattr(self, k, float(v))
                except ValueError:
                    pass


@AcquisitionFunctionConfig.register("simple_exploration")
class SimpleExplorationConfig(AcquisitionFunctionConfig):
    labels = {}
    defaults = "al_acquisition_function_simple_exploration"


@AcquisitionFunctionConfig.register("exploration_with_distance_penalty")
class ExplorationWithDistancePenaltyConfig(AcquisitionFunctionConfig):
    labels = {}
    defaults = "al_acquisition_function_exploration_with_distance_penalty"


@AcquisitionFunctionConfig.register("weighted_exploration")
class WeightedExplorationConfig(AcquisitionFunctionConfig):
    labels = {}
    defaults = "al_acquisition_function_weighted_exploration"


@AcquisitionFunctionConfig.register("probability_of_improvement")
class ProbabilityOfImprovementConfig(AcquisitionFunctionConfig):
    labels = {}
    defaults = "al_acquisition_function_probability_of_improvement"


@AcquisitionFunctionConfig.register("expected_improvement")
class ExpectedImprovementConfig(AcquisitionFunctionConfig):
    labels = {}
    defaults = "al_acquisition_function_expected_improvement"


@AcquisitionFunctionConfig.register("expected_improvement_2")
class ExpectedImprovement2Config(AcquisitionFunctionConfig):
    labels = {}
    defaults = "al_acquisition_function_expected_improvement_2"


@AcquisitionFunctionConfig.register("alternating_exploration")
class AlternatingExplorationConfig(AcquisitionFunctionConfig):
    labels = {}
    defaults = "al_acquisition_function_alternating_exploration"


@BaseConfig.register("ui")
class UIConfig(AbstractConfig):
    """Configuration for the Graphical User Interface."""
    labels = {}
    defaults = "ui"


@RunnerConfig.register("default")
@InterfaceConfig.register("default")
@PreConfig.register("default")
@PostConfig.register("default")
@AcquisitionFunctionConfig.register("default")
class DefaultConfig(AbstractConfig):
    """Default config for all run sub configs which just updates the attributes with user entries."""
    labels = {}
    defaults = None

    def __init__(self, **entries):
        name = entries.get('class', self.__class__.__name__)
        warnings.warn(f"Using default config for '{name}'.")
        self.update(**entries)

    def update(self, **entries):
        for name, value in entries.items():
            if hasattr(self, name) or name in map(str.lower, self.labels):
                attr = getattr(self, name, None)
                if isinstance(attr, dict):
                    attr.update(value)
                    setattr(self, name, attr)
                else:
                    setattr(self, name, value)
            else:
                setattr(self, name, value)



from profit.util.base_class import CustomABC
from numpy import ndarray
from typing import Union


class FileHandler(CustomABC):
    labels = {}
    associated_types = {'in': 'txt', 'out': 'txt'}

    @classmethod
    def save(cls, filename, data, **kwargs):
        """
        Parameters:
            filename (str)
            data (ndarray, dict)
            kwargs: Options like header and format for specific child classes.
        """

        ending = filename.split(".")[-1]
        if ending not in cls.labels:
            ending = cls.associated_types[ending]
        cls.labels[ending].save(filename, data, **kwargs)

    @classmethod
    def load(cls, filename, as_type='dtype'):
        """
        Parameters:
            filename (str)
            as_type (str): Identifier in which format the data should be returned.
                Options: dtype (structured array), dict
        """
        ending = filename.split(".")[-1]
        if ending not in cls.labels:
            ending = cls.associated_types[ending]
        return cls.labels[ending].load(filename, as_type)


@FileHandler.register("txt")
class TxtHandler(FileHandler):

    @classmethod
    def save(cls, filename, data, header=None, fmt=None):
        from numpy import hstack, savetxt
        try:
            if not header:
                header = " ".join(data.dtype.names)
            data = hstack([data[key] for key in data.dtype.names])
            if fmt:
                savetxt(filename, data, header=header, fmt=fmt)
            else:
                savetxt(filename, data, header=header)
        except TypeError:
            savetxt(filename, data)

    @classmethod
    def load(cls, filename, as_type='dtype'):
        from numpy import genfromtxt
        from profit.util.util import check_ndim
        names = True if as_type == 'dtype' else None
        return check_ndim(genfromtxt(filename, names=names))


@FileHandler.register("hdf5")
class HDF5Handler(FileHandler):

    @classmethod
    def save(cls, filename, data, **kwargs):
        from h5py import File
        with File(filename, 'w') as h5f:
            if hasattr(data, 'dtype'):
                # Save to numpy dtype names
                for key in data.dtype.names:
                    h5f[key] = data[key]
            elif isinstance(data, dict):
                # Save to dict key
                cls._recursive_dict2hdf(h5f, '', data)
            else:
                # Save to general data entry
                h5f['data'] = data

    @classmethod
    def load(cls, filename, as_type='dtype'):
        from h5py import File
        from numpy import array
        with File(filename, 'r') as h5f:
            if as_type == 'dtype':
                return cls.hdf2numpy(h5f)
            if as_type == 'dict':
                return cls.hdf2dict(h5f)
            else:
                return array(h5f['data'])

    @classmethod
    def _recursive_dict2hdf(cls, file, path, _dict):
        for key, value in _dict.items():
            if isinstance(value, dict):
                cls._recursive_dict2hdf(file, path + str(key) + '/', value)
            else:
                file[path + key] = value

    @staticmethod
    def hdf2numpy(dataset):
        from numpy import zeros_like, array
        from profit.util.util import check_ndim
        dtypes = [(key, float) for key in list(dataset.keys())]
        data = zeros_like(dataset[dtypes[0][0]], dtype=dtypes)
        for key in data.dtype.names:
            data[key] = array(dataset[key])
        return check_ndim(data)

    @staticmethod
    def hdf2dict(dataset):
        from numpy import array, ndarray, atleast_1d
        from h5py import Dataset
        load_dict = {}

        def recursive_hdf2dict(_data, _dict):
            for key in _data.keys():
                if isinstance(_data[key], Dataset):
                    val = _data[key][()]
                    if isinstance(val, bytes):  # Quick fix for new h5py version, which stores strings as bytes
                        val = val.decode('utf-8')
                    _dict[key] = atleast_1d(array(val)) if isinstance(val, ndarray) else val
                else:
                    _dict[key] = {}
                    recursive_hdf2dict(_data[key], _dict[key])

        recursive_hdf2dict(dataset, load_dict)
        return load_dict


@FileHandler.register("pkl")
class PickleHandler(FileHandler):

    @classmethod
    def save(cls, filename, data, **kwargs):
        from pickle import dump
        write_method = 'wb' if not 'method' in kwargs else kwargs['method']
        dump(data, open(filename, write_method))

    @classmethod
    def load(cls, filename, as_type='raw', read_method='rb'):
        from pickle import load
        if as_type != 'raw':
            return NotImplemented
        return load(open(filename, read_method))



"""

Explicit imports:
    profit.fit
    profit.run
    profit.sur
    profit.ui
"""

from . import main
from . import config
from . import util

# Retrieving package version at runtime
# https://pypi.org/project/setuptools-scm/
try:
    from setuptools_scm import get_version as version
    __version__ = version(root='..', relative_to=__file__)
    del version
except (ModuleNotFoundError, OSError):
    try:
        from importlib.metadata import version, PackageNotFoundError
    except ModuleNotFoundError:  # required for python3.7
        from importlib_metadata import version, PackageNotFoundError

    try:
        __version__ = version("profit")
    except PackageNotFoundError:
        # package is not installed
        __version__ = "?"

    del version, PackageNotFoundError  # clean up namespace



from abc import ABC

class CustomABC(ABC):

    labels = {}

    @classmethod
    def get_label(cls):
        """Returns the string label of a class object."""
        for label, item in cls.labels.items():
            if item == cls:
                return label
        raise NotImplementedError(f"Class {cls} is not implemented.")

    @classmethod
    def register(cls, label):
        """Decorator to register new classes."""
        def decorator(obj):
            if label in cls.labels:
                print(f"registering duplicate label '{label}' for {cls.__name__}.")
                #raise KeyError(f"registering duplicate label '{label}' for {cls.__name__}.")
            cls.labels[label] = obj
            return obj
        return decorator

    def __class_getitem__(cls, item):
        """Returns the child."""
        if item is None:
            return cls
        return cls.labels[item]
